"""The usbdevice commands module.

These commands are used in the following models:
DPO2K, DPO2KB, DPO4K, DPO4KB, MDO3, MDO3K, MDO4K, MDO4KB, MDO4KC, MSO2K, MSO2KB, MSO4K, MSO4KB

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - USBDevice:CONFigure {DISabled|IMAge|USBTmc}
    - USBDevice:CONFigure?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class UsbdeviceConfigure(SCPICmdWrite, SCPICmdRead):
    """The ``USBDevice:CONFigure`` command.

    **Description:**
        - Enables or disables the rear USB port for use with Pictbridge printers.

    **Usage:**
        - Using the ``.query()`` method will send the ``USBDevice:CONFigure?`` query.
        - Using the ``.verify(value)`` method will send the ``USBDevice:CONFigure?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``USBDevice:CONFigure value`` command.

    **SCPI Syntax:**

    ::

        - USBDevice:CONFigure {DISabled|IMAge|USBTmc}
        - USBDevice:CONFigure?

    **Info:**
        - ``DISabled`` disables the rear USB port.
        - ``IMAge`` enables the rear USB port as an SIC device.
        - ``USBTmc`` enables the rear USB port as a USBTMC device.
    """


class Usbdevice(SCPICmdRead):
    """The ``USBDevice`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``USBDevice?`` query.
        - Using the ``.verify(value)`` method will send the ``USBDevice?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.configure``: The ``USBDevice:CONFigure`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "USBDevice") -> None:
        super().__init__(device, cmd_syntax)
        self._configure = UsbdeviceConfigure(device, f"{self._cmd_syntax}:CONFigure")

    @property
    def configure(self) -> UsbdeviceConfigure:
        """Return the ``USBDevice:CONFigure`` command.

        **Description:**
            - Enables or disables the rear USB port for use with Pictbridge printers.

        **Usage:**
            - Using the ``.query()`` method will send the ``USBDevice:CONFigure?`` query.
            - Using the ``.verify(value)`` method will send the ``USBDevice:CONFigure?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``USBDevice:CONFigure value``
              command.

        **SCPI Syntax:**

        ::

            - USBDevice:CONFigure {DISabled|IMAge|USBTmc}
            - USBDevice:CONFigure?

        **Info:**
            - ``DISabled`` disables the rear USB port.
            - ``IMAge`` enables the rear USB port as an SIC device.
            - ``USBTmc`` enables the rear USB port as a USBTMC device.
        """
        return self._configure



"""The fpanel commands module.

These commands are used in the following models:
DPO2K, DPO2KB, MSO2K, MSO2KB

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - FPAnel:PRESS <button>
    - FPAnel:TURN <knob>,[<n>]
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class FpanelTurn(SCPICmdWrite):
    """The ``FPAnel:TURN`` command.

    **Description:**
        - Simulates the action of turning a specified front-panel control knob. When the front panel
          is locked, the front-panel button and multipurpose knob operations are suspended. The
          ``FPANEL:PRESS`` and commands will also not work, and, they will not generate an error.
          You can work around this by using the appropriate programmatic interface commands, instead
          of the front-panel commands. For example, to set the trigger level to 50%, you could use
          ``TRIGger:A SETLevel``. To force a trigger, you could use TRIGger FORCe.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``FPAnel:TURN value`` command.

    **SCPI Syntax:**

    ::

        - FPAnel:TURN <knob>,[<n>]

    **Info:**
        - ``<knob>`` is the name of a rotating control.
        - ``<n>`` represents the rotation direction and magnitude of rotation. Negative values
          represent a counterclockwise knob rotation, and positive values represent a clockwise
          rotation. The magnitude of <n> specifies the amount of the turn, where <n> = 1 represents
          turning the knob one unit, <n> = 2 represents turning the knob two units, <n> = 4
          represents turning the knob four units, and so on. The range of units depends on which
          front panel knob is specified.
        - ``Multipurpose a`` knob.
        - ``Multipurpose b`` knob.
        - ``Position`` knob.
        - ``Scale`` knob.
        - ``Level`` knob.
        - ``Position`` knob.
        - ``Scale`` knob.
    """


class FpanelPress(SCPICmdWrite):
    """The ``FPAnel:PRESS`` command.

    **Description:**
        - Simulates the action of pressing a specified front-panel button. When the front panel is
          locked, the front-panel button and multipurpose knob operations are suspended. The and the
          ``FPANEL:TURN`` commands will also not work. You can work around this by using the
          appropriate programmatic interface commands, instead of the front-panel commands.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``FPAnel:PRESS value`` command.

    **SCPI Syntax:**

    ::

        - FPAnel:PRESS <button>

    **Info:**
        - ``<button>`` is the name of a front-panel button. Most of the argument names associate
          directly with their front panel buttons. For example, AUTOSet is for the Autoset button.
        - ``Acquire`` button.
        - ``Arbitrary Function Generator`` button.
        - ``Autoset`` button.
        - ``Cursors`` button.
        - ``D<x>`` button x has a minimum of 15 and a maximum of 0.
        - ``Default Setup`` button.
        - ``Fine`` button.
        - ``Force Trig`` button.
        - ``Print`` button.
        - ``Intensity`` button.
        - ``Set/Clear`` button.
        - ``M`` button.
        - ``Menu Off`` button.
        - ``Measure`` button.
        - ``R`` button.
        - ``Run/Stop`` button.
        - ``Save`` button.
        - ``Menu`` button.
        - ``Search`` button.
        - ``Select`` button.
        - ``Single`` button.
        - ``Test`` button.
        - ``Menu`` button.
        - ``Utility`` button.
    """


class Fpanel(SCPICmdRead):
    """The ``FPAnel`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``FPAnel?`` query.
        - Using the ``.verify(value)`` method will send the ``FPAnel?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.press``: The ``FPAnel:PRESS`` command.
        - ``.turn``: The ``FPAnel:TURN`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "FPAnel") -> None:
        super().__init__(device, cmd_syntax)
        self._press = FpanelPress(device, f"{self._cmd_syntax}:PRESS")
        self._turn = FpanelTurn(device, f"{self._cmd_syntax}:TURN")

    @property
    def press(self) -> FpanelPress:
        """Return the ``FPAnel:PRESS`` command.

        **Description:**
            - Simulates the action of pressing a specified front-panel button. When the front panel
              is locked, the front-panel button and multipurpose knob operations are suspended. The
              and the ``FPANEL:TURN`` commands will also not work. You can work around this by using
              the appropriate programmatic interface commands, instead of the front-panel commands.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``FPAnel:PRESS value`` command.

        **SCPI Syntax:**

        ::

            - FPAnel:PRESS <button>

        **Info:**
            - ``<button>`` is the name of a front-panel button. Most of the argument names associate
              directly with their front panel buttons. For example, AUTOSet is for the Autoset
              button.
            - ``Acquire`` button.
            - ``Arbitrary Function Generator`` button.
            - ``Autoset`` button.
            - ``Cursors`` button.
            - ``D<x>`` button x has a minimum of 15 and a maximum of 0.
            - ``Default Setup`` button.
            - ``Fine`` button.
            - ``Force Trig`` button.
            - ``Print`` button.
            - ``Intensity`` button.
            - ``Set/Clear`` button.
            - ``M`` button.
            - ``Menu Off`` button.
            - ``Measure`` button.
            - ``R`` button.
            - ``Run/Stop`` button.
            - ``Save`` button.
            - ``Menu`` button.
            - ``Search`` button.
            - ``Select`` button.
            - ``Single`` button.
            - ``Test`` button.
            - ``Menu`` button.
            - ``Utility`` button.
        """
        return self._press

    @property
    def turn(self) -> FpanelTurn:
        """Return the ``FPAnel:TURN`` command.

        **Description:**
            - Simulates the action of turning a specified front-panel control knob. When the front
              panel is locked, the front-panel button and multipurpose knob operations are
              suspended. The ``FPANEL:PRESS`` and commands will also not work, and, they will not
              generate an error. You can work around this by using the appropriate programmatic
              interface commands, instead of the front-panel commands. For example, to set the
              trigger level to 50%, you could use ``TRIGger:A SETLevel``. To force a trigger, you
              could use TRIGger FORCe.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``FPAnel:TURN value`` command.

        **SCPI Syntax:**

        ::

            - FPAnel:TURN <knob>,[<n>]

        **Info:**
            - ``<knob>`` is the name of a rotating control.
            - ``<n>`` represents the rotation direction and magnitude of rotation. Negative values
              represent a counterclockwise knob rotation, and positive values represent a clockwise
              rotation. The magnitude of <n> specifies the amount of the turn, where <n> = 1
              represents turning the knob one unit, <n> = 2 represents turning the knob two units,
              <n> = 4 represents turning the knob four units, and so on. The range of units depends
              on which front panel knob is specified.
            - ``Multipurpose a`` knob.
            - ``Multipurpose b`` knob.
            - ``Position`` knob.
            - ``Scale`` knob.
            - ``Level`` knob.
            - ``Position`` knob.
            - ``Scale`` knob.
        """
        return self._turn



"""MSO70KC device driver module."""
from tm_devices.commands import MSO70KCMixin
from tm_devices.drivers.pi.scopes.tekscope_5k_7k_70k.mso70k import MSO70K


class MSO70KC(MSO70KCMixin, MSO70K):
    """MSO70KC device driver."""



"""SMU Model 2602B device driver module."""
from tm_devices.commands import SMU2602BMixin
from tm_devices.drivers.pi.source_measure_units.smu2600.smu2600b import SMU2600B


class SMU2602B(SMU2602BMixin, SMU2600B):
    """SMU2602B device driver."""



"""TekScope2k package init file."""
from tm_devices.drivers.pi.scopes.tekscope_2k.dpo2k import DPO2K
from tm_devices.drivers.pi.scopes.tekscope_2k.dpo2kb import DPO2KB
from tm_devices.drivers.pi.scopes.tekscope_2k.mso2k import MSO2K
from tm_devices.drivers.pi.scopes.tekscope_2k.mso2kb import MSO2KB
from tm_devices.drivers.pi.scopes.tekscope_2k.tekscope_2k import TekScope2k

__all__ = [
    "MSO2K",
    "MSO2KB",
    "DPO2K",
    "DPO2KB",
    "TekScope2k",
]



# pyright: reportUnnecessaryTypeIgnoreComment=none
"""Base Programmable Interface (PI) device driver module."""
import inspect
import os
import socket
import time

from abc import ABC, abstractmethod
from contextlib import contextmanager
from functools import cached_property
from typing import final, Generator, Optional, Sequence, Tuple, Union

import pyvisa as visa

from packaging.version import Version
from pyvisa import constants as visa_constants

from tm_devices.drivers.device import Device
from tm_devices.drivers.pi._ieee488_2_commands import IEEE4882Commands
from tm_devices.helpers import (
    check_visa_connection,
    create_visa_connection,
    DeviceConfigEntry,
    get_model_series,
    get_version,
    get_visa_backend,
    print_with_timestamp,
    PYVISA_PY_BACKEND,
)
from tm_devices.helpers.constants_and_dataclasses import UNIT_TEST_TIMEOUT


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class PIDevice(Device, ABC):
    """Base Programmable Interface (PI) device driver."""

    # This is a class constant that can be overwritten by children which defines
    # the class to use for the IEEE 488.2 commands.
    _IEEE_COMMANDS_CLASS = IEEE4882Commands

    ################################################################################################
    # Magic Methods
    ################################################################################################
    def __init__(
        self,
        config_entry: DeviceConfigEntry,
        verbose: bool,
        visa_resource: visa.resources.MessageBasedResource,
    ) -> None:
        """Create a PI device.

        Args:
            config_entry: A config entry object parsed by the DMConfigParser.
            verbose: A boolean indicating if verbose output should be printed.
            visa_resource: The VISA resource object.
        """
        super().__init__(config_entry, verbose)
        self._visa_resource: visa.resources.MessageBasedResource = visa_resource
        self._resource_expression = str(self._visa_resource.resource_info.resource_name)
        self._visa_library_path = self._visa_resource.visalib.library_path.path
        if self._visa_library_path == "py":  # pragma: no cover
            # For reboots to work, visa.ResourceManager needs "@py" passed in, not "py"
            self._visa_library_path = PYVISA_PY_BACKEND
        elif self._visa_library_path.endswith(".yaml"):  # pragma: no cover
            # Mark this as a simulated VISA backend
            self._visa_library_path += "@sim"
        self._visa_backend = self._get_visa_backend()
        # Use a default timeout of 30 seconds, if running unit tests use a smaller amount.
        self._default_visa_timeout = (
            30000
            if not bool(os.environ.get("TM_DEVICES_UNIT_TESTS_RUNNING"))
            else UNIT_TEST_TIMEOUT
        )
        self._ieee_cmds = self._IEEE_COMMANDS_CLASS(self)  # type: ignore
        self.reset_visa_timeout()

    ################################################################################################
    # Abstract Properties
    ################################################################################################
    @property
    @abstractmethod
    def all_channel_names_list(self) -> Tuple[str, ...]:
        """Return a tuple containing all the channel names."""

    @cached_property
    @abstractmethod
    def total_channels(self) -> int:
        """Return the total number of channels (all types)."""

    ################################################################################################
    # Abstract Methods
    ################################################################################################
    @abstractmethod
    def expect_esr(self, esr: Union[int, str], error_string: str = "") -> Tuple[bool, str]:
        r"""Check for the expected number of errors and output string.

        Args:
            esr: Expected ``*ESR?`` value
            error_string: Expected error buffer string.
                Multiple errors should be separated by a \n character

        Returns:
            Boolean indicating if the check passed or failed and a string with the results.
        """

    @abstractmethod
    def get_eventlog_status(self) -> Tuple[bool, str]:
        """Help function for getting the eventlog status.

        Returns:
            Boolean indicating no error, String containing concatenated contents of event log.
        """

    def turn_channel_off(self, channel_str: str) -> None:
        """Turn off the specified channel.

        Args:
            channel_str: The name of the channel to turn off.

        Raises:
            NotImplementedError: Indicates the current driver has not implemented this method.
        """
        # TODO: implement for all driver subclasses then remove this blanket NotImplementedError
        raise NotImplementedError(
            f"``.{inspect.currentframe().f_code.co_name}()``"  # pyright: ignore
            f" is not yet implemented for the {self.__class__.__name__} driver"
        )

    def turn_channel_on(self, channel_str: str) -> None:
        """Turn on the specified channel.

        Args:
            channel_str: The name of the channel to turn on.

        Raises:
            NotImplementedError: Indicates the current driver has not implemented this method.
        """
        # TODO: implement for all driver subclasses then remove this blanket NotImplementedError
        raise NotImplementedError(
            f"``.{inspect.currentframe().f_code.co_name}()``"  # pyright: ignore
            f" is not yet implemented for the {self.__class__.__name__} driver"
        )

    ################################################################################################
    # Properties
    ################################################################################################
    @property
    def default_visa_timeout(self) -> int:
        """Return the default VISA timeout value in milliseconds."""
        return self._default_visa_timeout

    @property
    def ieee_cmds(self) -> IEEE4882Commands:
        """Return an internal class containing methods for the standard IEEE 488.2 command set."""
        return self._ieee_cmds

    @property
    def resource_expression(self) -> str:
        """Return the VISA resource expression."""
        return self._resource_expression

    @property
    def visa_backend(self) -> str:
        """Return the VISA backend in use."""
        return self._visa_backend

    @property
    def visa_timeout(self) -> float:
        """Return the current VISA timeout of the device in milliseconds."""
        return self._visa_resource.timeout

    @visa_timeout.setter
    def visa_timeout(self, timeout_ms: float) -> None:
        """Set the VISA timeout of the device in milliseconds.

        Args:
            timeout_ms: The new VISA timeout, in milliseconds.
        """
        self._visa_resource.timeout = timeout_ms
        if self._verbose:
            print(f"{self._name_and_alias} VISA timeout set to: {timeout_ms}ms")

    @property
    def visa_resource(self) -> visa.resources.MessageBasedResource:
        """Return the VISA resource object.

        This gives access to all the attributes and methods that PyVISA provides.
        """
        return self._visa_resource

    ################################################################################################
    # Cached Properties
    ################################################################################################
    @cached_property
    def sw_version(self) -> Version:
        """Return the software version of the device."""
        id_string_parts = self.idn_string.split(",")
        sw_version = id_string_parts[-1]
        if ":" in id_string_parts[-1]:
            sw_version = id_string_parts[-1].rsplit(":", 1)[-1]
        if "-" in sw_version:
            split_sw = sw_version.split("-")
            retval = get_version(split_sw[0])
        elif " " in sw_version.lstrip():
            split_sw = sw_version.split(" ")
            retval = get_version(split_sw[0])
        else:
            retval = get_version(sw_version)
        return retval

    @cached_property
    def idn_string(self) -> str:
        r"""Return the string returned from the ``*IDN?`` query when the device was created."""
        return self.ieee_cmds.idn()

    @cached_property
    def manufacturer(self) -> str:
        """Return the manufacturer of the device."""
        return self.idn_string.split(",")[0].strip()

    @cached_property
    def model(self) -> str:
        """Return the full model of the device."""
        return self.idn_string.split(",")[1].strip()

    @cached_property
    def serial(self) -> str:
        """Return the serial number of the device."""
        return self.idn_string.split(",")[2].strip()

    @cached_property
    def series(self) -> str:
        """Return the series of the device.

        Returns:
            The series of the device, e.g. MSO5, TSOVu, TEKSCOPE, AFG3K, AWG5200
        """
        return get_model_series(self.model)

    ################################################################################################
    # Context Manager Methods
    ################################################################################################
    @contextmanager
    def temporary_visa_timeout(self, temporary_timeout_ms: float) -> Generator[None, None, None]:
        """Set a temporary VISA timeout value for the duration of the context.

        This will reset the VISA timeout to the previous value when the context exits.

        Args:
            temporary_timeout_ms: The temporary VISA timeout value, in milliseconds.
        """
        old_timeout = self.visa_timeout
        self.visa_timeout = temporary_timeout_ms
        try:
            yield
        finally:
            self.visa_timeout = old_timeout

    ################################################################################################
    # Public Methods
    ################################################################################################
    @final
    def check_visa_connection(self, verbose: bool = True) -> bool:
        """Check if a VISA connection can be made to the device.

        Wrapper function for :py:func:`~tm_devices.helpers.check_visa_connection`.

        Args:
            verbose: Set this to False in order to disable printouts.
        """
        return check_visa_connection(
            self._config_entry,
            self._visa_library_path,
            self._name_and_alias,
            verbose=verbose and self._verbose,
        )

    def device_clear(self) -> None:  # pragma: no cover
        """Clear the input and output buffers of the VISA device."""
        self._visa_resource.clear()

    def disable_srq_events(self) -> None:  # pragma: no cover
        """Disable the service request event for the device."""
        self._visa_resource.disable_event(
            visa_constants.VI_EVENT_SERVICE_REQ, visa_constants.VI_QUEUE  # type: ignore
        )

    def enable_srq_events(self) -> None:  # pragma: no cover
        """Enable the service request event for the device."""
        self._visa_resource.enable_event(
            visa_constants.VI_EVENT_SERVICE_REQ, visa_constants.VI_QUEUE  # type: ignore
        )

    def get_visa_stb(self) -> int:  # pragma: no cover
        """Return the VISA status byte."""
        return self._visa_resource.read_stb()

    def query(
        self,
        query: str,
        *,
        verbose: bool = True,
        remove_quotes: bool = False,
        allow_empty: bool = False,
    ) -> str:
        """Send a query to the device and return the result.

        Args:
            query: The query to send to the device.
            verbose: Set this to False in order to disable printouts.
            remove_quotes: Set this to True to remove all double quotes from the returned value.
            allow_empty: Set this to True if an empty return string is permitted.

        Returns:
            The results of the query.

        Raises:
            Error: An error occurred while sending the command.
            SystemError: An empty string was returned from the device.
        """
        if self._verbose and verbose:
            print_with_timestamp(f"({self._name_and_alias}) Query >>  {query!r}")

        try:
            response = self._visa_resource.query(query).strip()
            if remove_quotes:
                response = response.replace('"', "")
        except (visa.VisaIOError, socket.error) as error:
            pi_cmd_repr = f" for {query!r} " if self._verbose and verbose else " "
            msg = f"The query{pi_cmd_repr}failed with the following message: {error!r}"
            raise visa.Error(msg) from error

        if self._verbose and verbose:
            print_with_timestamp(f"Response from {query!r} >>  {response!r}")

        if not allow_empty and not response:
            pi_cmd_repr = (
                f" for the following query: {query!r} " if self._verbose and verbose else ""
            )
            msg = f"An empty string was returned{pi_cmd_repr}"
            raise SystemError(msg)

        return response

    def query_binary(self, query: str, verbose: bool = True) -> Sequence[float]:
        """Send a query to the device and return the binary values.

        Args:
            query: The query to send to the device.
            verbose: Set this to False in order to disable printouts.

        Returns:
            The results of the query.

        Raises:
            Error: An error occurred while sending the command.
            SystemError: An empty string was returned from the device.
        """
        if self._verbose and verbose:
            print_with_timestamp(f"({self._name_and_alias}) Query Binary Values >>  {query!r}")

        try:
            response = self._visa_resource.query_binary_values(query)  # pyright: ignore
        except (visa.VisaIOError, socket.error) as error:
            pi_cmd_repr = f" for {query!r} " if self._verbose and verbose else " "
            msg = f"The query{pi_cmd_repr}failed with the following message: {error!r}"
            raise visa.Error(msg) from error

        if self._verbose and verbose:
            print_with_timestamp(
                f"Response from {query!r} >>  {', '.join([str(x) for x in response])!r}"
            )

        if not response:
            pi_cmd_repr = (
                f" for the following query: {query!r} " if self._verbose and verbose else ""
            )
            msg = f"An empty string was returned{pi_cmd_repr}"
            raise SystemError(msg)

        return response

    def query_expect_timeout(
        self, invalid_command: str, timeout_ms: int = 200, verbose: bool = True
    ) -> str:
        """Send a query expecting it to time out and fail.

        Args:
            invalid_command:
                The bad command to query.
            timeout_ms:
                The VISA timeout duration for this function (in milliseconds).
            verbose:
                A flag to decide if we should print out a message indicating what query was sent.

        Returns:
            The timeout error or the response to the query.

        Raises:
            visa.errors.Error: Any non-timeout error will be re-raised.
        """
        retval = ""
        with self.temporary_visa_timeout(timeout_ms):
            try:
                retval = self.query(invalid_command, verbose=verbose)
            except visa.errors.Error as error:
                error_args = error.args[0]
                # Only want to gracefully handle a timeout
                if (
                    "VisaIOError('VI_ERROR_TMO" in error_args
                    and "Timeout expired before operation completed.')" in error_args
                ):
                    # VISA function ends the current query interaction (probably buffer related)
                    self.device_clear()
                else:
                    raise
        return retval

    def query_less_than(
        self,
        query: str,
        value: Union[float, str],
        tolerance: float = 0,
        percentage: bool = False,
        allow_equal: bool = False,
    ) -> bool:
        """Send the given query and verify the result is less than the expected response.

        Args:
            query: The command that is being checked.
            value: The expected value of the query.
            tolerance: The acceptable difference between two floating point values.
            percentage: A boolean indicating what kind of tolerance check to perform.
                 False means absolute tolerance: +/- tolerance.
                 True means percent tolerance: +/- (tolerance / 100) * value.
            allow_equal: A boolean indicating if equal results should be allowed.

        Returns:
            Boolean indicating whether the check passed or failed.
        """
        query_passed = True
        actual_value = float(self.query(query).strip())
        tolerance = float(tolerance)
        value = float(value)
        if percentage:
            tolerance = abs((tolerance / 100.0) * value)
        # Verify that the number is within the tolerance.
        # Also check to make sure that the string of each number isn't
        # identical, this prevents issues from returned values that have
        # a trailing zero or some other non-contributing character that
        # will cause comparison issues.
        if (
            allow_equal
            and (
                (not abs(actual_value) <= (abs(value) + tolerance))
                and (str(value) != str(actual_value))
            )
        ) or (not allow_equal and not abs(actual_value) < (abs(value) + tolerance)):
            max_value = value + tolerance
            self.raise_failure(
                f"query_less_than failed for query: {query}\n  "
                f"max ({'inclusive' if allow_equal else 'exclusive'}): "
                f"{max_value}\n  act: {actual_value}"
            )

        return query_passed

    def query_raw_binary(self, query: str, verbose: bool = True) -> bytes:
        """Send a command to the device and then read and return the raw binary values.

        Args:
            query: The query to send to the device.
            verbose: Set this to False in order to disable printouts.

        Returns:
            The raw results of the query.

        Raises:
            Error: An error occurred while sending the command.
            SystemError: An empty string was returned from the device.
        """
        if self._verbose and verbose:
            print_with_timestamp(f"({self._name_and_alias}) Query Raw Binary >>  {query!r}")

        try:
            self._visa_resource.write(query)
            response = self.read_raw()
        except (visa.VisaIOError, socket.error) as error:
            pi_cmd_repr = f" for {query!r} " if self._verbose and verbose else " "
            msg = f"The query{pi_cmd_repr}failed with the following message: {error!r}"
            raise visa.Error(msg) from error

        if self._verbose and verbose:
            print_with_timestamp(f"Response from {query!r} >>  {response!r}")

        if not response.strip():
            pi_cmd_repr = (
                f" for the following query: {query!r} " if self._verbose and verbose else ""
            )
            msg = f"An empty string was returned{pi_cmd_repr}"
            raise SystemError(msg)

        return response

    def query_response(
        self,
        query: str,
        value: Union[str, float],
        tolerance: float = 0,
        percentage: bool = False,
        remove_quotes: bool = False,
        custom_message_prefix: str = "",
    ) -> Tuple[bool, str]:
        """Query the and verify the result.

        Args:
            query: The query that is being checked.
            value: The expected value of the query.
            tolerance: The acceptable difference between two floating point values.
            percentage: A boolean indicating what kind of tolerance check to perform.
                 False means absolute tolerance: +/- tolerance.
                 True means percent tolerance: +/- (tolerance / 100) * value.
            remove_quotes: Set this to True to remove all double quotes from the returned value.
            custom_message_prefix: A custom message to be prepended to the failure message.

        Returns:
            Tuple containing the boolean verification result and the value returned from the query.
        """
        actual_value = self.query(query, remove_quotes=remove_quotes)
        message_prefix = f"query_response failed for query: {query}"
        if custom_message_prefix:
            message_prefix = f"{custom_message_prefix}\n{message_prefix}"
        query_passed = self.verify_values(
            value,
            actual_value,
            tolerance=tolerance,
            percentage=percentage,
            custom_message_prefix=message_prefix,
        )
        return query_passed, actual_value

    def query_response_not(
        self,
        query: str,
        value: str,
        remove_quotes: bool = False,
        custom_message_prefix: str = "",
    ) -> Tuple[bool, str]:
        """Query the and verify the result is not the given value.

        Args:
            query: The query that is being checked.
            value: The value that the query should not return.
            remove_quotes: Set this to True to remove all double quotes from the returned value.
            custom_message_prefix: A custom message to be prepended to the failure message.

        Returns:
            Tuple containing the boolean verification result and the value returned from the query.
        """
        actual_value = self.query(query, remove_quotes=remove_quotes)
        message_prefix = f"query_response_not failed for query: {query}"
        if custom_message_prefix:
            message_prefix = f"{custom_message_prefix}\n{message_prefix}"
        query_passed = self.verify_values(
            value,
            actual_value,
            custom_message_prefix=message_prefix,
            expect_fail=True,
        )
        return query_passed, actual_value

    def read(self) -> str:
        """Return the read results from the VISA resource."""
        return self._visa_resource.read()

    def read_raw(self) -> bytes:
        """Return the read_raw results from the VISA resource."""
        return self._visa_resource.read_raw()

    def reset_visa_timeout(self) -> None:
        """Reset the VISA timeout to the default value."""
        self.visa_timeout = self._default_visa_timeout

    def set_and_check(  # noqa: PLR0913
        self,
        command: str,
        value: Union[str, float],
        tolerance: float = 0,
        percentage: bool = False,
        remove_quotes: bool = False,
        custom_message_prefix: str = "",
        *,
        expected_value: Optional[Union[str, float]] = None,
    ) -> str:
        """Send the given command with the given value and then verify the results.

        Args:
            command: The command to send.
                For example: ``:AFG:FUNCTION``
            value: The value being set by the command.
                For example: ``SINE``
            tolerance: The acceptable difference between two floating point values.
            percentage: A boolean indicating what kind of tolerance check to perform.
                 False means absolute tolerance: +/- tolerance.
                 True means percent tolerance: +/- (tolerance / 100) * value.
            remove_quotes: Set this to True to remove all double quotes from the returned value.
            custom_message_prefix: A custom message to be prepended to the failure message.
            expected_value: An optional, alternative value expected to be returned.

        Returns:
            The output of the query portion of the method.
        """
        self.write(f"{command} {value}")
        if self._enable_verification:
            check = self.query(command + "?", remove_quotes=remove_quotes)
            message_prefix = f"Failed to set {command} to {value}"
            if custom_message_prefix:
                message_prefix = f"{custom_message_prefix}\n{message_prefix}"
            self.verify_values(
                value if expected_value is None else expected_value,
                check,
                tolerance=tolerance,
                percentage=percentage,
                custom_message_prefix=message_prefix,
                log_error=True,
            )
        else:
            check = ""
        return check

    @final
    def turn_all_channels_off(self) -> None:
        """Turn all channels off."""
        for ch_str in self.all_channel_names_list:
            self.turn_channel_off(ch_str)

    @final
    def turn_all_channels_on(self) -> None:
        """Turn all channels on."""
        for ch_str in self.all_channel_names_list:
            self.turn_channel_on(ch_str)

    def wait_for_srq_event(self, timeout: int) -> visa.resources.resource.WaitResponse:
        """Wait for the service request event to happen, up to the given timeout.

        Args:
            timeout: Time to wait (in seconds).

        Returns:
            The WaitResponse for the SRQ event.
        """
        # The timeout value is multiplied by 1000 because the function expects milliseconds.
        return self._visa_resource.wait_on_event(
            visa_constants.VI_EVENT_SERVICE_REQ,  # type: ignore
            timeout * 1000,
        )

    def wait_for_visa_connection(
        self,
        wait_time: float,
        sleep_seconds: int = 5,
        accept_immediate_connection: bool = False,
        verbose: bool = True,
    ) -> bool:
        """Wait for a VISA connection to be made to the device.

        Args:
            wait_time: The maximum time to wait in seconds.
            sleep_seconds: The number of seconds to sleep in between connection attempts.
            accept_immediate_connection: A boolean indicating if a connection on the
                                         first attempt is a valid connection.
            verbose: Set this to False in order to disable printouts.

        Returns:
            A boolean indicating if a VISA connection was made within the given time limit.

        Raises:
            AssertionError: Indicating that the device erroneously connected on the first try.
        """
        attempt_num = 0
        visa_connection = False
        if verbose:
            print_with_timestamp(
                f"Attempting to establish a VISA connection with {self._resource_expression}"
            )
        start_time = time.perf_counter()
        while (time.perf_counter() - start_time) <= wait_time:
            if visa_connection := self.check_visa_connection(verbose=False):
                # pylint: disable=compare-to-zero
                if not (attempt_num == 0 and not accept_immediate_connection):
                    break
                msg = (
                    f"A VISA connection was established with "
                    f"{self._resource_expression} on the first attempt "
                    f"when it should not have been."
                )
                raise AssertionError(msg)
            time.sleep(sleep_seconds)
            attempt_num += 1

        end_time = time.perf_counter()
        total_time = end_time - start_time

        if verbose:
            if visa_connection:
                print_with_timestamp(
                    f"Successfully established a VISA connection with {self._resource_expression} "
                    f"after {total_time:.2f} seconds"
                )
            else:
                print_with_timestamp(
                    f"Unable to establish a VISA connection with {self._resource_expression} "
                    f"after {total_time:.2f} seconds"
                )

        return visa_connection

    def write(self, command: str, opc: bool = False, verbose: bool = True) -> None:
        r"""Write a command to the device.

        Args:
            command: The command to write to the device
            opc: Boolean indicating if ``*OPC?`` should be queried after sending the command.
            verbose: Set this to False in order to disable printouts.

        Raises:
            Error: An error occurred while sending the command.
            SystemError: ``*OPC?`` did not return "1" after sending the command.
        """
        if self._verbose and verbose:
            print_with_timestamp(f"({self._name_and_alias}) Write >>  {command!r}")

        try:
            self._visa_resource.write(command)
        except (visa.VisaIOError, socket.error) as error:
            pi_cmd_repr = f" for {command!r} " if self._verbose and verbose else " "
            msg = f"The write{pi_cmd_repr}failed with the following message: {error!r}"
            raise visa.Error(msg) from error

        if opc and (result := self.ieee_cmds.opc()) != "1":
            pi_cmd_repr = f" {command!r}" if self._verbose and verbose else " the command"
            msg = f"After issuing{pi_cmd_repr}, OPC returned incorrect data: {result!r}"
            raise SystemError(msg)

    def write_raw(self, command: bytes, verbose: bool = True) -> None:
        r"""Write a raw command to the device.

        Args:
            command: The command to write to the device
            verbose: Set this to False in order to disable printouts.

        Raises:
            Error: An error occurred while sending the command.
        """
        if self._verbose and verbose:
            print_with_timestamp(f"({self._name_and_alias}) Write Raw >>  {command!r}")

        try:
            self._visa_resource.write_raw(command)
        except (visa.VisaIOError, socket.error) as error:
            pi_cmd_repr = f" for {command!r} " if self._verbose and verbose else " "
            msg = f"The raw write{pi_cmd_repr}failed with the following message: {error!r}"
            raise visa.Error(msg) from error

    ################################################################################################
    # Public Methods (PI Library)
    ################################################################################################
    def factory_reset(self) -> None:
        r"""Send the ``FACTORY`` command followed by an ``*OPC?`` query.

        This command is equivalent to pressing the DEFAULT SETUP button located on
        the instrument front panel or selected Default Setup from the File menu.

        It recalls the instrument to factory default settings.

        In addition to what ``*RST`` does, this command also performs the following operations:

        - Clears any pending OPC operations
        - Resets the following IEEE488.2 registers:
        - ``*ESR`` 0 (Event Status Enable Register)
        - ``*SRE`` 0 (Service Request Enable Register)
        - DESE 255 (Device Event Status Enable Register)
        - ``*PSC`` 1 (Power-on Status Clear Flag)
        - Deletes all defined aliases.
        - Enables command headers (:HEADer 1).

        ``FACTORY`` only resets the programmable interface settings, it does not change the
        user interface settings.
        """
        self.write("FACTORY", opc=True)

    def reset(self) -> None:
        r"""Send the ``*RST`` command followed by an ``*OPC?`` query.

        ``*RST`` only resets the programmable interface settings, it does not change the user
        interface settings.
        """
        self.ieee_cmds.rst()
        self.ieee_cmds.opc()

    ################################################################################################
    # Private Methods
    ################################################################################################
    def _cleanup(self) -> None:
        """Perform the cleanup defined for the device."""
        self.visa_timeout = (
            120000
            if not bool(os.environ.get("TM_DEVICES_UNIT_TESTS_RUNNING"))
            else UNIT_TEST_TIMEOUT
        )
        self.ieee_cmds.cls()
        self.reset()
        self.reset_visa_timeout()
        self.expect_esr(0)

    def _close(self) -> None:
        """Close this device and all its used resources and components."""
        self._visa_resource.close()
        self._visa_resource = None  # type: ignore
        self._is_open = False

    def _get_visa_backend(self) -> str:
        """Determine what the VISA backend is for this device.

        Returns:
            A string containing the VISA backend.
        """
        return get_visa_backend(self._visa_resource.visalib.library_path.path)

    def _has_errors(self) -> bool:
        """Check if the device has any errors.

        Returns:
            A boolean indicating if any errors were found in the device.
        """
        return not self.expect_esr(0)[0]

    def _open(self) -> bool:
        """Open necessary resources and components and return a boolean indicating success."""
        opened = True
        if self._visa_resource is None:  # type: ignore
            opened = False
            # 5 seconds when running unit tests, else 600 seconds (10 minutes)
            num_seconds_to_attempt_reconnection = (
                5 if bool(os.environ.get("TM_DEVICES_UNIT_TESTS_RUNNING")) else 600
            )
            sleep_time_seconds = 5
            # Attempt to reconnect with the device many times.
            for iteration in range(num_seconds_to_attempt_reconnection // sleep_time_seconds):
                try:
                    time.sleep(sleep_time_seconds)
                    # Create new VISA connection
                    self._visa_resource = create_visa_connection(
                        self._config_entry,
                        visa_library=self._visa_library_path,
                    )
                    opened = True
                    if iteration < 1 and not bool(
                        os.environ.get("TM_DEVICES_UNIT_TESTS_REBOOT_ALLOW")
                    ):
                        # It shouldn't be able to reconnect on the first try.
                        # If it does, it most certainly did not reboot.
                        opened = False
                    break
                except ConnectionError:
                    # raised by the create_visa_connection() function
                    pass

        self._is_open = opened
        return opened



# ruff: noqa: D402,PLR0913
# pyright: reportConstantRedefinition=none
"""The digio commands module.

These commands are used in the following models:
SMU2606B, SMU2651A, SMU2657A

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Attributes and Functions:

::

    - digio.readbit()
    - digio.readport()
    - digio.trigger[N].assert()
    - digio.trigger[N].clear()
    - digio.trigger[N].mode
    - digio.trigger[N].overrun
    - digio.trigger[N].pulsewidth
    - digio.trigger[N].release()
    - digio.trigger[N].reset()
    - digio.trigger[N].stimulus
    - digio.trigger[N].wait()
    - digio.writebit()
    - digio.writeport()
    - digio.writeprotect
"""
from typing import Dict, Optional, TYPE_CHECKING, Union

from .._helpers import (
    BaseTSPCmd,
    DefaultDictPassKeyToFactory,
    NoDeviceProvidedError,
    ValidatedDynamicNumberCmd,
)

if TYPE_CHECKING:
    from tm_devices.drivers.pi.tsp_device import TSPDevice


class DigioTriggerItem(ValidatedDynamicNumberCmd, BaseTSPCmd):
    """The ``digio.trigger[N]`` command tree.

    **Info:**
        - ``N``, the digital I/O trigger line (1 to 14).

    Constants:
        - ``.EVENT_ID``: The trigger event generated by the digital I/O line N.

    Properties/methods:
        - ``.assert()``: The ``digio.trigger[N].assert()`` function.
        - ``.clear()``: The ``digio.trigger[N].clear()`` function.
        - ``.mode``: The ``digio.trigger[N].mode`` attribute.
        - ``.overrun``: The ``digio.trigger[N].overrun`` attribute.
        - ``.pulsewidth``: The ``digio.trigger[N].pulsewidth`` attribute.
        - ``.release()``: The ``digio.trigger[N].release()`` function.
        - ``.reset()``: The ``digio.trigger[N].reset()`` function.
        - ``.stimulus``: The ``digio.trigger[N].stimulus`` attribute.
        - ``.wait()``: The ``digio.trigger[N].wait()`` function.
    """

    EVENT_ID = "digio.trigger[N].EVENT_ID"
    """str: The trigger event generated by the digital I/O line N."""

    def __init__(self, device: Optional["TSPDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        # pylint: disable=invalid-name
        self.EVENT_ID = self.EVENT_ID.replace(
            "[N]", f"[{self._cmd_syntax.rsplit('[', maxsplit=1)[-1].split(']', maxsplit=1)[0]}]"
        )

    @property
    def mode(self) -> str:
        """Access the ``digio.trigger[N].mode`` attribute.

        **Description:**
            - This attribute sets the mode in which the trigger event detector and the output
              trigger generator operate on the given trigger line.

        **Usage:**
            - Accessing this property will send the ``print(digio.trigger[N].mode)`` query.
            - Setting this property to a value will send the ``digio.trigger[N].mode = value``
              command.

        **TSP Syntax:**

        ::

            - digio.trigger[N].mode = value
            - print(digio.trigger[N].mode)

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".mode"
            return self._device.query(f"print({self._cmd_syntax}.mode)")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.mode`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @mode.setter
    def mode(self, value: Union[str, float]) -> None:
        """Access the ``digio.trigger[N].mode`` attribute.

        **Description:**
            - This attribute sets the mode in which the trigger event detector and the output
              trigger generator operate on the given trigger line.

        **Usage:**
            - Accessing this property will send the ``print(digio.trigger[N].mode)`` query.
            - Setting this property to a value will send the ``digio.trigger[N].mode = value``
              command.

        **TSP Syntax:**

        ::

            - digio.trigger[N].mode = value
            - print(digio.trigger[N].mode)

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_verification_enabled:  # type: ignore[union-attr]
                self._device.set_and_check(  # type: ignore[union-attr]
                    self._cmd_syntax + ".mode", value
                )
            else:
                self._device.write(f"{self._cmd_syntax}.mode = {value}")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.mode`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @property
    def overrun(self) -> str:
        """Access the ``digio.trigger[N].overrun`` attribute.

        **Description:**
            - This attribute returns the event detector overrun status.

        **Usage:**
            - Accessing this property will send the ``print(digio.trigger[N].overrun)`` query.

        **TSP Syntax:**

        ::

            - print(digio.trigger[N].overrun)

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".overrun"
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.overrun)"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.overrun`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @property
    def pulsewidth(self) -> str:
        """Access the ``digio.trigger[N].pulsewidth`` attribute.

        **Description:**
            - This attribute describes the length of time that the trigger line is asserted for
              output triggers.

        **Usage:**
            - Accessing this property will send the ``print(digio.trigger[N].pulsewidth)`` query.
            - Setting this property to a value will send the ``digio.trigger[N].pulsewidth = value``
              command.

        **TSP Syntax:**

        ::

            - digio.trigger[N].pulsewidth = value
            - print(digio.trigger[N].pulsewidth)

        **Info:**
            - ``width``, the pulse width (seconds).
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".pulsewidth"
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.pulsewidth)"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.pulsewidth`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @pulsewidth.setter
    def pulsewidth(self, value: Union[str, float]) -> None:
        """Access the ``digio.trigger[N].pulsewidth`` attribute.

        **Description:**
            - This attribute describes the length of time that the trigger line is asserted for
              output triggers.

        **Usage:**
            - Accessing this property will send the ``print(digio.trigger[N].pulsewidth)`` query.
            - Setting this property to a value will send the ``digio.trigger[N].pulsewidth = value``
              command.

        **TSP Syntax:**

        ::

            - digio.trigger[N].pulsewidth = value
            - print(digio.trigger[N].pulsewidth)

        **Info:**
            - ``width``, the pulse width (seconds).
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_verification_enabled:  # type: ignore[union-attr]
                self._device.set_and_check(  # type: ignore[union-attr]
                    self._cmd_syntax + ".pulsewidth", value
                )
            else:
                self._device.write(  # type: ignore[union-attr]
                    f"{self._cmd_syntax}.pulsewidth = {value}"
                )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.pulsewidth`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @property
    def stimulus(self) -> str:
        """Access the ``digio.trigger[N].stimulus`` attribute.

        **Description:**
            - This attribute selects the event that causes a trigger to be asserted on the digital
              output line.

        **Usage:**
            - Accessing this property will send the ``print(digio.trigger[N].stimulus)`` query.
            - Setting this property to a value will send the ``digio.trigger[N].stimulus = value``
              command.

        **TSP Syntax:**

        ::

            - digio.trigger[N].stimulus = value
            - print(digio.trigger[N].stimulus)

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".stimulus"
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.stimulus)"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.stimulus`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @stimulus.setter
    def stimulus(self, value: Union[str, float]) -> None:
        """Access the ``digio.trigger[N].stimulus`` attribute.

        **Description:**
            - This attribute selects the event that causes a trigger to be asserted on the digital
              output line.

        **Usage:**
            - Accessing this property will send the ``print(digio.trigger[N].stimulus)`` query.
            - Setting this property to a value will send the ``digio.trigger[N].stimulus = value``
              command.

        **TSP Syntax:**

        ::

            - digio.trigger[N].stimulus = value
            - print(digio.trigger[N].stimulus)

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_verification_enabled:  # type: ignore[union-attr]
                self._device.set_and_check(  # type: ignore[union-attr]
                    self._cmd_syntax + ".stimulus", value
                )
            else:
                self._device.write(  # type: ignore[union-attr]
                    f"{self._cmd_syntax}.stimulus = {value}"
                )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.stimulus`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def assert_(self) -> None:
        """Run the ``digio.trigger[N].assert()`` function.

        **Description:**
            - This function asserts a trigger pulse on one of the digital I/O lines.

        **TSP Syntax:**

        ::

            - digio.trigger[N].assert()

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(f"{self._cmd_syntax}.assert()")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.assert()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def clear(self) -> None:
        """Run the ``digio.trigger[N].clear()`` function.

        **Description:**
            - This function clears the trigger event on a digital I/O line.

        **TSP Syntax:**

        ::

            - digio.trigger[N].clear()

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(f"{self._cmd_syntax}.clear()")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.clear()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def release(self) -> None:
        """Run the ``digio.trigger[N].release()`` function.

        **Description:**
            - This function releases an indefinite length or latched trigger.

        **TSP Syntax:**

        ::

            - digio.trigger[N].release()

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(f"{self._cmd_syntax}.release()")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.release()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def reset(self) -> None:
        """Run the ``digio.trigger[N].reset()`` function.

        **Description:**
            - This function resets trigger values to their factory defaults.

        **TSP Syntax:**

        ::

            - digio.trigger[N].reset()

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(f"{self._cmd_syntax}.reset()")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.reset()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def wait(self, timeout: float) -> str:
        """Run the ``digio.trigger[N].wait()`` function.

        **Description:**
            - This function waits for a trigger.

        **TSP Syntax:**

        ::

            - digio.trigger[N].wait()

        Args:
            timeout: Timeout in seconds.

        Returns:
            The result of the function call.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.wait({timeout}))"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.wait()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error


class Digio(BaseTSPCmd):
    """The ``digio`` command tree.

    Constants:
        - ``.TRIG_BYPASS``: Sets the mode in which the trigger event detector and the output trigger
          generator operate on the specified trigger line to allow direct control of the line.
        - ``.TRIG_EITHER``: Sets the mode in which the trigger event detector and the output trigger
          generator operate on the specified trigger line to detect rising- or falling-edge triggers
          as input and assert a TTL-low pulse for output.
        - ``.TRIG_FALLING``: Sets the mode in which the trigger event detector and the output
          trigger generator operate on the specified trigger line to detect falling-edge triggers as
          input and assert a TTL-low pulse for output.
        - ``.TRIG_RISING``: Sets the mode in which the trigger event detector and the output trigger
          generator operate on the specified trigger line so that if the programmed state of the
          line is high, the digio.TRIG_RISING mode behavior is similar to digio.TRIG_RISINGA. If the
          programmed state of the line is low, the digio.TRIG_RISING mode behavior is similar to
          digio.TRIG_RISINGM. Only use this setting if necessary for compatibility with other
          Keithley Instruments products.
        - ``.TRIG_RISINGA``: Sets the mode in which the trigger event detector and the output
          trigger generator operate on the specified trigger line to detect rising-edge triggers as
          input and assert a TTL-low pulse for output.
        - ``.TRIG_RISINGM``: Sets the mode in which the trigger event detector and the output
          trigger generator operate on the specified trigger line to assert a TTL-high pulse for
          output. Input edge detection is not possible in this mode.
        - ``.TRIG_SYNCHRONOUS``: Sets the mode in which the trigger event detector and the output
          trigger generator operate on the specified trigger line to detect the falling-edge input
          triggers and automatically latch and drive the trigger line low. Asserts a TTL-low pulse
          as an output trigger.
        - ``.TRIG_SYNCHRONOUSA``: Sets the mode in which the trigger event detector and the output
          trigger generator operate on the specified trigger line to detect the falling-edge input
          triggers and automatically latch and drive the trigger line low. Asserting the output
          trigger releases the latched line.
        - ``.TRIG_SYNCHRONOUSM``: Sets the mode in which the trigger event detector and the output
          trigger generator operate on the specified trigger line to detect rising-edge triggers as
          input and assert a TTL-low pulse for output.

    Properties/methods:
        - ``.readbit()``: The ``digio.readbit()`` function.
        - ``.readport()``: The ``digio.readport()`` function.
        - ``.trigger``: The ``digio.trigger[N]`` command tree.
        - ``.writebit()``: The ``digio.writebit()`` function.
        - ``.writeport()``: The ``digio.writeport()`` function.
        - ``.writeprotect``: The ``digio.writeprotect`` attribute.
    """

    TRIG_BYPASS = "digio.TRIG_BYPASS"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to allow direct control of the line."""  # noqa: E501
    TRIG_EITHER = "digio.TRIG_EITHER"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to detect rising- or falling-edge triggers as input and assert a TTL-low pulse for output."""  # noqa: E501
    TRIG_FALLING = "digio.TRIG_FALLING"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to detect falling-edge triggers as input and assert a TTL-low pulse for output."""  # noqa: E501
    TRIG_RISING = "digio.TRIG_RISING"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line so that if the programmed state of the line is high, the digio.TRIG_RISING mode behavior is similar to digio.TRIG_RISINGA. If the programmed state of the line is low, the digio.TRIG_RISING mode behavior is similar to digio.TRIG_RISINGM. Only use this setting if necessary for compatibility with other Keithley Instruments products."""  # noqa: E501
    TRIG_RISINGA = "digio.TRIG_RISINGA"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to detect rising-edge triggers as input and assert a TTL-low pulse for output."""  # noqa: E501
    TRIG_RISINGM = "digio.TRIG_RISINGM"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to assert a TTL-high pulse for output. Input edge detection is not possible in this mode."""  # noqa: E501
    TRIG_SYNCHRONOUS = "digio.TRIG_SYNCHRONOUS"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to detect the falling-edge input triggers and automatically latch and drive the trigger line low. Asserts a TTL-low pulse as an output trigger."""  # noqa: E501
    TRIG_SYNCHRONOUSA = "digio.TRIG_SYNCHRONOUSA"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to detect the falling-edge input triggers and automatically latch and drive the trigger line low. Asserting the output trigger releases the latched line."""  # noqa: E501
    TRIG_SYNCHRONOUSM = "digio.TRIG_SYNCHRONOUSM"
    """str: Sets the mode in which the trigger event detector and the output trigger generator operate on the specified trigger line to detect rising-edge triggers as input and assert a TTL-low pulse for output."""  # noqa: E501

    def __init__(self, device: Optional["TSPDevice"] = None, cmd_syntax: str = "digio") -> None:
        super().__init__(device, cmd_syntax)
        self._trigger: Dict[int, DigioTriggerItem] = DefaultDictPassKeyToFactory(
            lambda x: DigioTriggerItem(device, f"{self._cmd_syntax}.trigger[{x}]")
        )

    @property
    def trigger(self) -> Dict[int, DigioTriggerItem]:
        """Return the ``digio.trigger[N]`` command tree.

        **Info:**
            - ``N``, the digital I/O trigger line (1 to 14).

        Constants:
            - ``.EVENT_ID``: The trigger event generated by the digital I/O line N.

        Sub-properties/methods:
            - ``.assert()``: The ``digio.trigger[N].assert()`` function.
            - ``.clear()``: The ``digio.trigger[N].clear()`` function.
            - ``.mode``: The ``digio.trigger[N].mode`` attribute.
            - ``.overrun``: The ``digio.trigger[N].overrun`` attribute.
            - ``.pulsewidth``: The ``digio.trigger[N].pulsewidth`` attribute.
            - ``.release()``: The ``digio.trigger[N].release()`` function.
            - ``.reset()``: The ``digio.trigger[N].reset()`` function.
            - ``.stimulus``: The ``digio.trigger[N].stimulus`` attribute.
            - ``.wait()``: The ``digio.trigger[N].wait()`` function.
        """
        return self._trigger

    @property
    def writeprotect(self) -> str:
        """Access the ``digio.writeprotect`` attribute.

        **Description:**
            - This attribute contains the write-protect mask that protects bits from changes from
              the digio.writebit() and digio.writeport() functions.

        **Usage:**
            - Accessing this property will send the ``print(digio.writeprotect)`` query.
            - Setting this property to a value will send the ``digio.writeprotect = value`` command.

        **TSP Syntax:**

        ::

            - digio.writeprotect = value
            - print(digio.writeprotect)

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".writeprotect"
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.writeprotect)"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.writeprotect`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @writeprotect.setter
    def writeprotect(self, value: Union[str, float]) -> None:
        """Access the ``digio.writeprotect`` attribute.

        **Description:**
            - This attribute contains the write-protect mask that protects bits from changes from
              the digio.writebit() and digio.writeport() functions.

        **Usage:**
            - Accessing this property will send the ``print(digio.writeprotect)`` query.
            - Setting this property to a value will send the ``digio.writeprotect = value`` command.

        **TSP Syntax:**

        ::

            - digio.writeprotect = value
            - print(digio.writeprotect)

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_verification_enabled:  # type: ignore[union-attr]
                self._device.set_and_check(  # type: ignore[union-attr]
                    self._cmd_syntax + ".writeprotect", value
                )
            else:
                self._device.write(  # type: ignore[union-attr]
                    f"{self._cmd_syntax}.writeprotect = {value}"
                )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.writeprotect`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def readbit(self, n: int) -> str:
        """Run the ``digio.readbit()`` function.

        **Description:**
            - This function reads one digital I/O line.

        **TSP Syntax:**

        ::

            - digio.readbit()

        Args:
            n: Digital I/O line number to be read (1 to 14).

        Returns:
            The result of the function call.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.readbit({n}))"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.readbit()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def readport(self) -> str:
        """Run the ``digio.readport()`` function.

        **Description:**
            - This function reads the digital I/O port.

        **TSP Syntax:**

        ::

            - digio.readport()

        Returns:
            The result of the function call.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.readport())"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.readport()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def writebit(self, n: int, data: int) -> None:
        """Run the ``digio.writebit()`` function.

        **Description:**
            - This function sets a digital I/O line high or low.

        **TSP Syntax:**

        ::

            - digio.writebit()

        Args:
            n: Digital I/O trigger line (1 to 14).
            data: The value to write to the bit.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(  # type: ignore[union-attr]
                f"{self._cmd_syntax}.writebit({n}, {data})"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.writebit()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def writeport(self, data: int) -> None:
        """Run the ``digio.writeport()`` function.

        **Description:**
            - This function writes to all digital I/O lines.

        **TSP Syntax:**

        ::

            - digio.writeport()

        Args:
            data: Value to write to the port (0 to 16383).

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(f"{self._cmd_syntax}.writeport({data})")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.writeport()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error



"""Config parser module."""
from __future__ import annotations

import contextlib
import dataclasses
import os
import pathlib

from types import MappingProxyType
from typing import Any, cast, Dict, get_type_hints, List, Mapping, Optional, Tuple, Type, Union

import tomli
import tomli_w
import yaml

from tm_devices.helpers import (
    AliasDict,
    ConfigFileType,
    ConnectionTypes,
    DeviceConfigEntry,
    DeviceTypes,
    DMConfigOptions,
    SerialConfig,
)
from tm_devices.helpers.constants_and_dataclasses import (
    CONFIG_CLASS_STR_PREFIX_MAPPING,
    DataclassProtocol,
)


class DMConfigParser:
    """Class to parse a configuration file for device connections and runtime options."""

    # This is the environment variable that we check for a user-provided,
    # comma-delimited list of device parameters, multiple devices are delimited with "~~~".
    DEVICES_ENV_VARIABLE = "TM_DEVICES"
    # comma-delimited list of options.
    OPTIONS_ENV_VARIABLE = "TM_OPTIONS"
    # This is the environment variable that we check for a user-provided path to a custom config
    # file that, if present, should contain a list of devices and hostnames/IP addresses.
    CONFIG_FILE_PATH_ENV_VARIABLE = "TM_DEVICES_CONFIG"
    # The default location for a config file that, if present, should contain a list of devices and
    # hostnames/IP addresses.
    DEFAULT_CONFIG_FILE_PATH = "./tm_devices.yaml"
    # Easy access to the enum class
    FileType = ConfigFileType

    _CONFIG_NESTED_DICT_MAPPING: Mapping[
        Union[Type[DataclassProtocol], Type[SerialConfig]], str
    ] = MappingProxyType({SerialConfig: "serial_config"})

    ################################################################################################
    # Magic Methods
    ################################################################################################
    def __init__(self) -> None:
        """Initialize the DMConfigParser and read the configuration."""
        # Create a count of the number of each type of device created
        self.__dev_count: Dict[str, int] = {dev: 0 for dev in DeviceTypes.list_values()}

        self.__devices: Dict[str, DeviceConfigEntry] = AliasDict()
        self.__options = DMConfigOptions()

        # Check if the environment variables exists
        if self.DEVICES_ENV_VARIABLE in os.environ or self.OPTIONS_ENV_VARIABLE in os.environ:
            devices_list = self.__parse_env_devices()
            self.update_options(self.__parse_env_options())

            self.__pre_process_env_devices(devices_list)
            self.__add_from_device_list(devices_list)
        else:
            # Errors can be ignored, no default file was found
            with contextlib.suppress(FileNotFoundError):
                self.load_config_file()

    def __str__(self) -> str:
        """Return string representation of instance as environment variables."""
        return (
            f"{self.OPTIONS_ENV_VARIABLE}={self.__options}\n"
            f"{self.DEVICES_ENV_VARIABLE}="
            f"{'~~~'.join( ['', *map(str, self.__devices.values()), ''])}"
        )

    ################################################################################################
    # Properties
    ################################################################################################
    @property
    def defined_config_file_path(self) -> str:
        """Filepath of the config file.

        Prioritizes path defined from env variable ```self.CONFIG_FILE_PATH_ENV_VARIABLE``` with a
        fallback to the default path (`tm_devices.yaml` in runtime cwd directory).
        """
        return os.getenv(
            self.CONFIG_FILE_PATH_ENV_VARIABLE,
            self.DEFAULT_CONFIG_FILE_PATH,
        )

    @property
    def devices(self) -> Mapping[str, DeviceConfigEntry]:
        """Return the device dict as a read-only object."""
        return MappingProxyType(self.__devices)

    @property
    def options(self) -> DMConfigOptions:
        """Return the config options."""
        return self.__options

    ################################################################################################
    # Public Methods
    ################################################################################################
    def add_device(  # noqa: PLR0913
        self,
        *,
        device_type: Union[DeviceTypes, str],
        address: str,
        connection_type: Union[ConnectionTypes, str] = ConnectionTypes.TCPIP,
        alias: Optional[str] = None,
        lan_port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
        device_driver: Optional[str] = None,
    ) -> Tuple[str, DeviceConfigEntry]:
        """Add a new device configuration entry.

        Args:
            device_type: The specific device type defined in the config entry.
            address: The address used to connect to the device:
                TCPIP: IP address or the hostname.
                SOCKET: IP address or the hostname (must define `lan_port` field).
                REST_API: IP address or the hostname (must define `lan_port` field).
                SERIAL/ASRL: serial COM port number.
                USB: use expression format f"{model}-{serial_number}" (ex: "MSO24-ABC0123").
            connection_type: The specific type of connection defined in the config entry.
            alias: An optional key/name used to retrieve this device from the DeviceManager.
            lan_port: The port number to connect on, used for SOCKET/REST_API connections.
            serial_config: A dataclass for holding serial connection info.
            device_driver: A string indicating the specific Python device driver to use.

        Returns:
            Tuple of the config entry name and the new DeviceConfigEntry

        Raises:
            ValueError: invalid config options, trouble creating (or a duplicate) device entry.
        """
        device_type = cast(DeviceTypes, device_type)
        connection_type = cast(ConnectionTypes, connection_type)
        # device and connection type conversion to enum from string is handled on initialization
        new_entry = DeviceConfigEntry(
            device_type=device_type,
            address=address,
            connection_type=connection_type,
            alias=alias,
            lan_port=lan_port,
            serial_config=serial_config,
            device_driver=device_driver,
        )
        # Currently, USB connections when using the PyVISA-py backend are not supported.
        if self.__options.standalone and new_entry.connection_type in {
            ConnectionTypes.USB,
            ConnectionTypes.GPIB,
        }:
            message = (
                f"{new_entry.connection_type.value} connections using PyVISA-py (standalone) "
                f"are not supported.\n{new_entry}"
            )
            raise ValueError(message)

        # Validate the connection is unique
        for dev_entry in self.__devices.values():
            if new_entry.get_address_expression() == dev_entry.get_address_expression():
                message = (
                    f"Found duplicate addresses in the "
                    f'configuration for "{new_entry.get_address_expression()}":'
                    f"\n\t    Entry 1: {dev_entry}\n\tNew Entry 2: {new_entry}"
                )
                raise ValueError(message)

        # increment count
        self.__dev_count[new_entry.device_type.value] += 1
        device_number = self.__dev_count[new_entry.device_type.value]
        new_entry_name = f"{new_entry.device_type.value} {device_number}"
        # store the new entry
        self.__devices[new_entry_name] = new_entry
        if new_entry.alias:
            self.__devices[new_entry.alias] = new_entry
        return new_entry_name, new_entry

    def load_config_file(
        self, config_file_path: Optional[Union[str, os.PathLike[str]]] = None
    ) -> None:
        """Load in the config file located at the given path.

        This method will update any changed options and add any newly defined devices.

        Args:
            config_file_path: The path to the config file to load.
        """
        # Grabs path from environment variable with fallback to the default path
        file_path = config_file_path if config_file_path else self.defined_config_file_path
        # Automatically handles toml and yaml
        options, devices_list = self.__parse_config_file(file_path)
        self.__add_from_device_list(devices_list)
        self.update_options(options)

    def remove_device(self, device_name: str) -> None:
        """Remove a device from the configuration.

        Args:
            device_name: The name of the device to remove from the internal dictionary.
        """
        del self.__devices[device_name]

    def to_config_file_text(self, file_type: ConfigFileType = ConfigFileType.YAML) -> str:
        """Return string representation of instance as config file contents.

        Args:
            file_type: The type of file to create, yaml or toml.
        """
        # order of dict matters to match auto-formatted toml and yaml
        if file_type == self.FileType.TOML:
            retval = tomli_w.dumps(
                {
                    "devices": [dev.to_dict(ignore_none=True) for dev in self.devices.values()],
                    "options": self.options.to_dict(ignore_none=True),
                }
            )
        else:
            retval = yaml.dump(
                {
                    "options": self.options.to_dict(ignore_none=True),
                    "devices": [dev.to_dict(ignore_none=True) for dev in self.devices.values()],
                }
            )
            # add document header
            retval = "---\n" + retval
            # add extra spacing under devices section
            devs_idx = retval.find("devices:") + 8
            opts_idx = retval.find("options:") - 1
            retval = retval.replace(
                retval[devs_idx:opts_idx], retval[devs_idx:opts_idx].replace("\n", "\n  ")
            )
        return retval

    def update_options(self, new_options: DMConfigOptions) -> None:
        """Update the current configuration options.

        This ensures that any time new options are loaded in only explicitly defined options will be
        updated.

        Args:
            new_options: The new options to merge with the current options.
        """
        self.__options = dataclasses.replace(
            self.__options,
            **new_options.to_dict(ignore_none=True),
        )

    def write_config_to_file(
        self, config_file_path: Optional[Union[str, os.PathLike[str]]] = None
    ) -> str:
        """Write a config file located at the current working directory (or custom path).

        This method will overwrite any existing config file with the current devices and options.
        If no custom path is provided, and the ``TM_DEVICES_CONFIG`` environment variable is not set
        defaults to a yaml file in the current working directory.

        Args:
            config_file_path: The path to the config file. If ends in ".toml" will create toml file.

        Returns:
            Path to the written file.
        """
        if config_file_path is None:
            config_file_path = self.defined_config_file_path
        else:
            config_file_path = cast(str, config_file_path)

        with open(config_file_path, "w", encoding="utf-8") as config_file:
            file_type = (
                DMConfigParser.FileType.TOML
                if config_file_path.lower().endswith(".toml")
                else DMConfigParser.FileType.YAML
            )
            config_file.write(self.to_config_file_text(file_type=file_type))

        return config_file_path

    ################################################################################################
    # Private Methods
    ################################################################################################
    def __add_from_device_list(self, devices_list: List[Dict[str, Any]]) -> None:
        """Add devices from a list to the current config.

        Args:
            devices_list: The list of devices to add to the current config.
        """
        for entry in devices_list:
            # Instantiate nested classes from fields that store kwargs-style dictionaries
            for to_class, entry_key in self._CONFIG_NESTED_DICT_MAPPING.items():
                if config_dict := entry.get(entry_key):
                    # noinspection PyArgumentList
                    entry[entry_key] = to_class(**config_dict)
            self.add_device(**entry)

    def __parse_env_options(self) -> DMConfigOptions:
        """Extract the options flags from the env variable TM_OPTIONS.

        Returns:
            Dataclass of config options/flags.

        Raises:
            KeyError: Indicates unrecognized option name.
        """
        options_list = [
            arg.strip() for arg in os.getenv(self.OPTIONS_ENV_VARIABLE, "").split(",") if arg
        ]
        valid_config_options = {option.upper() for option in get_type_hints(DMConfigOptions)}
        if valid_config_options.issuperset(options_list):
            options = {
                arg_name.lower(): arg_name in options_list for arg_name in valid_config_options
            }
            return DMConfigOptions(**options)
        msg = (
            f"Invalid configuration options found: {list(set(options_list) - valid_config_options)}"
        )
        raise KeyError(msg)

    def __parse_env_devices(self) -> List[Dict[str, Any]]:
        """Extract list of device configuration dictionaries from the env variable TM_DEVICES.

        Returns:
            List of device entry dictionaries.

        Raises:
            ValueError: invalid string formatting cannot be parsed.
        """
        devices_str_list = [
            arg.strip() for arg in os.getenv(self.DEVICES_ENV_VARIABLE, "").split("~~~") if arg
        ]
        retval: List[Dict[str, Any]] = []
        msg: List[str] = []
        for dev_entry in devices_str_list:
            temp_dict = {}
            for dev_arg in dev_entry.split(","):
                try:
                    dev_key, dev_val = dev_arg.split("=", 1)
                    temp_dict[dev_key] = dev_val
                except ValueError:  # noqa: PERF203
                    msg.append(
                        f'Expected device defined like "device_type=SCOPE,address=1.2.3.4,...",\n'
                        f'cannot parse "{dev_entry}"'
                    )
            retval.append(temp_dict)
        if msg:
            raise ValueError("\n".join(msg))
        return retval

    @staticmethod
    def __parse_config_file(
        config_file_path: Union[str, os.PathLike[str]]
    ) -> Tuple[DMConfigOptions, List[Dict[str, Any]]]:
        """Parse config file for the options flags and list of device configuration dictionaries.

        Args:
            config_file_path: The path to the config file to load.

        Returns:
            Tuple with dataclass of config options/flags and list of device entry dictionaries.

        Raises:
            FileNotFoundError: if the defined config file does not exist.
            KeyError: Indicates unrecognized option name.
        """
        config_path = pathlib.Path(config_file_path)  # normalize the path
        if not os.path.isfile(config_path):
            raise FileNotFoundError(config_path)
        # read in data
        with open(config_path, encoding="utf-8") as config_file:
            if config_path.suffix == ".toml":
                data = tomli.loads(config_file.read())
            else:  # ["yaml", "yml"]
                data = yaml.safe_load(config_file)
        # return data with some pre-processing
        options_mapping = data.get("options", {})
        try:
            options = DMConfigOptions(**options_mapping)
        except TypeError:
            options_list = list(options_mapping)
            valid_options = set(get_type_hints(DMConfigOptions))
            msg = f"Invalid configuration options found: {list(set(options_list) - valid_options)}"
            raise KeyError(msg)  # noqa: TRY200,B904
        devices_list = data.get("devices", [])
        return options, devices_list

    def __pre_process_env_devices(self, devices_list: List[Dict[str, Any]]) -> None:
        """Modify device entries in place to make env data mimic toml/yaml parser output format."""
        for entry in devices_list:
            # Bundle any prefixed pairs into a nested kwarg-style dict.
            for to_class, prefix in CONFIG_CLASS_STR_PREFIX_MAPPING.items():
                config_dict: Dict[str, str]
                if config_dict := {
                    key.replace(prefix, ""): entry.pop(key)
                    for key in list(entry.keys())
                    if key.startswith(prefix)
                    and (key not in self._CONFIG_NESTED_DICT_MAPPING.values())
                }:
                    # assign the dict to the correct key via the class the prefix represented
                    entry[self._CONFIG_NESTED_DICT_MAPPING[to_class]] = config_dict



# ruff: noqa: D402,PLR0913
# pyright: reportConstantRedefinition=none
"""The localnode commands module.

These commands are used in the following models:
SMU2601B_Pulse

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Attributes and Functions:

::

    - localnode.model
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import BaseTSPCmd, NoDeviceProvidedError

if TYPE_CHECKING:
    from tm_devices.drivers.pi.tsp_device import TSPDevice


class Localnode(BaseTSPCmd):
    """The ``localnode`` command tree.

    Constants:
        - ``.PASSWORD_ALL``: Use passwords on the web interface and all remote command interfaces.
        - ``.PASSWORD_LAN``: Use passwords on the web interface and all LAN interfaces.
        - ``.PASSWORD_NONE``: Disable passwords everywhere.
        - ``.PASSWORD_WEB``: Use passwords on the web interface only.

    Properties/methods:
        - ``.model``: The ``localnode.model`` attribute.
    """

    PASSWORD_ALL = "localnode.PASSWORD_ALL"
    """str: Use passwords on the web interface and all remote command interfaces."""
    PASSWORD_LAN = "localnode.PASSWORD_LAN"
    """str: Use passwords on the web interface and all LAN interfaces."""
    PASSWORD_NONE = "localnode.PASSWORD_NONE"
    """str: Disable passwords everywhere."""
    PASSWORD_WEB = "localnode.PASSWORD_WEB"
    """str: Use passwords on the web interface only."""

    def __init__(self, device: Optional["TSPDevice"] = None, cmd_syntax: str = "localnode") -> None:
        super().__init__(device, cmd_syntax)

    @property
    def model(self) -> str:
        """Access the ``localnode.model`` attribute.

        **Description:**
            - This attribute stores the model number.

        **Usage:**
            - Accessing this property will send the ``print(localnode.model)`` query.

        **TSP Syntax:**

        ::

            - print(localnode.model)

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".model"
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.model)"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.model`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error



"""TSOVu device driver module."""
import inspect

import pyvisa as visa

from tm_devices.drivers.device import family_base_class
from tm_devices.drivers.pi.scopes.scope import Scope
from tm_devices.helpers import DeviceConfigEntry


@family_base_class
class TSOVu(Scope):
    """TSOVu device driver."""

    ################################################################################################
    # Magic Methods
    ################################################################################################
    def __init__(
        self,
        config_entry: DeviceConfigEntry,
        verbose: bool,
        visa_resource: visa.resources.MessageBasedResource,
    ) -> None:
        """Create a TSOVu device.

        Args:
            config_entry: A config entry object parsed by the DMConfigParser.
            verbose: A boolean indicating if verbose output should be printed.
            visa_resource: The VISA resource object.
        """
        super().__init__(config_entry, verbose, visa_resource)
        self.write("HEADER OFF", verbose=False)

    ################################################################################################
    # Private Methods
    ################################################################################################
    def _reboot(self) -> None:
        """Perform the actual rebooting code.

        Raises:
            NotImplementedError: Indicates the current driver has not implemented this method.
        """
        # TODO: implement
        raise NotImplementedError(
            f"``.{inspect.currentframe().f_code.co_name}()``"  # pyright: ignore
            f" is not yet implemented for the {self.__class__.__name__} driver"
        )



"""The ref commands module.

These commands are used in the following models:
DPO5KB, DPO70KC, DPO70KD, DPO70KDX, DPO70KSX, DPO7KC, DSA70KC, DSA70KD, MSO5KB, MSO70KC, MSO70KDX

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - REF<x>:HORizontal:POSition <NR3>
    - REF<x>:HORizontal:POSition?
    - REF<x>:LABel:NAMe <QString>
    - REF<x>:LABel:NAMe?
    - REF<x>:LABel:XPOS <NR1>
    - REF<x>:LABel:XPOS?
    - REF<x>:LABel:YPOS <NR1>
    - REF<x>:LABel:YPOS?
    - REF<x>:LABel?
    - REF<x>:THRESHold <NR3>
    - REF<x>:THRESHold?
    - REF<x>:VERTical:POSition <NR3>
    - REF<x>:VERTical:POSition?
    - REF<x>:VERTical:SCAle <NR3>
    - REF<x>:VERTical:SCAle?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite, ValidatedDynamicNumberCmd

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class RefItemVerticalScale(SCPICmdWrite, SCPICmdRead):
    """The ``REF<x>:VERTical:SCAle`` command.

    **Description:**
        - This command specifies the vertical scale for the reference waveform specified by <x>.
          This setting controls the display only. For a signal with constant amplitude, increasing
          the scale causes the waveform to be displayed smaller. Decreasing the scale causes the
          waveform to be displayed larger.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:VERTical:SCAle?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:VERTical:SCAle?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``REF<x>:VERTical:SCAle value`` command.

    **SCPI Syntax:**

    ::

        - REF<x>:VERTical:SCAle <NR3>
        - REF<x>:VERTical:SCAle?

    **Info:**
        - ``<NR3>`` is a floating point number that specifies the gain in user units-per-division.
    """


class RefItemVerticalPosition(SCPICmdWrite, SCPICmdRead):
    """The ``REF<x>:VERTical:POSition`` command.

    **Description:**
        - This command specifies the vertical position of the reference waveform specified by <x>.
          Increasing the position value of a waveform causes the waveform to move up, and decreasing
          the position value causes the waveform to move down. Position adjusts only the display
          position of a waveform. The position value determines the vertical graticule coordinate at
          which signal values are displayed. For example, if the position for Reference 3 is set to
          2.0, the signal represented by that reference will be displayed at 2.0 divisions above the
          center of the screen.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:VERTical:POSition?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:VERTical:POSition?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``REF<x>:VERTical:POSition value``
          command.

    **SCPI Syntax:**

    ::

        - REF<x>:VERTical:POSition <NR3>
        - REF<x>:VERTical:POSition?

    **Info:**
        - ``<NR3>`` is a floating point number that specifies the desired position, in divisions
          from the center horizontal graticule. The range is from -5.0 to 5.0 divisions.
    """


class RefItemVertical(SCPICmdRead):
    """The ``REF<x>:VERTical`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:VERTical?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:VERTical?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.position``: The ``REF<x>:VERTical:POSition`` command.
        - ``.scale``: The ``REF<x>:VERTical:SCAle`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._position = RefItemVerticalPosition(device, f"{self._cmd_syntax}:POSition")
        self._scale = RefItemVerticalScale(device, f"{self._cmd_syntax}:SCAle")

    @property
    def position(self) -> RefItemVerticalPosition:
        """Return the ``REF<x>:VERTical:POSition`` command.

        **Description:**
            - This command specifies the vertical position of the reference waveform specified by
              <x>. Increasing the position value of a waveform causes the waveform to move up, and
              decreasing the position value causes the waveform to move down. Position adjusts only
              the display position of a waveform. The position value determines the vertical
              graticule coordinate at which signal values are displayed. For example, if the
              position for Reference 3 is set to 2.0, the signal represented by that reference will
              be displayed at 2.0 divisions above the center of the screen.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:VERTical:POSition?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:VERTical:POSition?`` query
              and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``REF<x>:VERTical:POSition value``
              command.

        **SCPI Syntax:**

        ::

            - REF<x>:VERTical:POSition <NR3>
            - REF<x>:VERTical:POSition?

        **Info:**
            - ``<NR3>`` is a floating point number that specifies the desired position, in divisions
              from the center horizontal graticule. The range is from -5.0 to 5.0 divisions.
        """
        return self._position

    @property
    def scale(self) -> RefItemVerticalScale:
        """Return the ``REF<x>:VERTical:SCAle`` command.

        **Description:**
            - This command specifies the vertical scale for the reference waveform specified by <x>.
              This setting controls the display only. For a signal with constant amplitude,
              increasing the scale causes the waveform to be displayed smaller. Decreasing the scale
              causes the waveform to be displayed larger.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:VERTical:SCAle?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:VERTical:SCAle?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``REF<x>:VERTical:SCAle value``
              command.

        **SCPI Syntax:**

        ::

            - REF<x>:VERTical:SCAle <NR3>
            - REF<x>:VERTical:SCAle?

        **Info:**
            - ``<NR3>`` is a floating point number that specifies the gain in user
              units-per-division.
        """
        return self._scale


class RefItemThreshold(SCPICmdWrite, SCPICmdRead):
    """The ``REF<x>:THRESHold`` command.

    **Description:**
        - This command sets or queries the comparable threshold for converting the reference signal
          to digital form for the channel specified by x. The value of x can range from 1 through 4.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:THRESHold?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:THRESHold?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``REF<x>:THRESHold value`` command.

    **SCPI Syntax:**

    ::

        - REF<x>:THRESHold <NR3>
        - REF<x>:THRESHold?

    **Info:**
        - ``<NR3>`` specifies the reference threshold in volts.
    """


class RefItemLabelYpos(SCPICmdWrite, SCPICmdRead):
    """The ``REF<x>:LABel:YPOS`` command.

    **Description:**
        - This command sets or queries the Y screen offset at which the label (attached to the
          displayed waveform of the specified reference) is displayed, relative to the center of the
          screen. The reference waveform is specified by x. The value of x can range from 1 through
          4. This command is equivalent to selecting Reference Waveforms from the File menu,
          choosing Label from the drop-down list, selecting the tab associated with the reference
          for which you want to position a label, and entering a value in the Y Position box.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:LABel:YPOS?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:LABel:YPOS?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``REF<x>:LABel:YPOS value`` command.

    **SCPI Syntax:**

    ::

        - REF<x>:LABel:YPOS <NR1>
        - REF<x>:LABel:YPOS?

    **Info:**
        - ``<NR1>`` is the location (in divisions) where the waveform label for the selected
          reference is displayed, relative to the waveform handle. Arguments should be integers
          ranging from 10 to -10.
    """


class RefItemLabelXpos(SCPICmdWrite, SCPICmdRead):
    """The ``REF<x>:LABel:XPOS`` command.

    **Description:**
        - This command sets or queries the X screen offset at which the label (attached to the
          displayed waveform of the specified reference) is displayed, relative to the left edge of
          the screen The reference waveform is specified by x. The value of x can range from 1
          through 4. This command is equivalent to selecting Reference Waveforms from the File menu,
          choosing Label from the drop-down list, selecting the tab associated with the reference
          for which you want to position a label, and entering a value in the X Position box.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:LABel:XPOS?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:LABel:XPOS?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``REF<x>:LABel:XPOS value`` command.

    **SCPI Syntax:**

    ::

        - REF<x>:LABel:XPOS <NR1>
        - REF<x>:LABel:XPOS?

    **Info:**
        - ``<NR1>`` is the location (control in divisions) where the waveform label for the selected
          reference is displayed, relative to the left edge of the screen. Arguments should be
          integers ranging from 0 through 10.
    """


class RefItemLabelName(SCPICmdWrite, SCPICmdRead):
    """The ``REF<x>:LABel:NAMe`` command.

    **Description:**
        - This command sets or queries the label of the designated waveform. The reference waveform
          is specified by x. The value of x can range from 1 through 4. This command is equivalent
          to selecting Reference Waveforms from the File menu, choosing Label from the drop-down
          list, selecting the tab associated with the reference for which you want to create a
          label, and entering a label in the Label box.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:LABel:NAMe?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:LABel:NAMe?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``REF<x>:LABel:NAMe value`` command.

    **SCPI Syntax:**

    ::

        - REF<x>:LABel:NAMe <QString>
        - REF<x>:LABel:NAMe?

    **Info:**
        - ``<QString>`` is the character string that will be used for the reference waveform label
          name.
    """

    _WRAP_ARG_WITH_QUOTES = True


class RefItemLabel(SCPICmdRead):
    """The ``REF<x>:LABel`` command.

    **Description:**
        - This query-only command returns a branch query containing the waveform label name and the
          coordinates at which the label (attached to the displayed waveform of the specified
          reference) is displayed. The reference waveform is specified by x. The value of x can
          range from 1 through 4. This command is equivalent to selecting Reference Waveforms from
          the File menu and then choosing Label from the drop-down list.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:LABel?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:LABel?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    **SCPI Syntax:**

    ::

        - REF<x>:LABel?

    Properties:
        - ``.name``: The ``REF<x>:LABel:NAMe`` command.
        - ``.xpos``: The ``REF<x>:LABel:XPOS`` command.
        - ``.ypos``: The ``REF<x>:LABel:YPOS`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._name = RefItemLabelName(device, f"{self._cmd_syntax}:NAMe")
        self._xpos = RefItemLabelXpos(device, f"{self._cmd_syntax}:XPOS")
        self._ypos = RefItemLabelYpos(device, f"{self._cmd_syntax}:YPOS")

    @property
    def name(self) -> RefItemLabelName:
        """Return the ``REF<x>:LABel:NAMe`` command.

        **Description:**
            - This command sets or queries the label of the designated waveform. The reference
              waveform is specified by x. The value of x can range from 1 through 4. This command is
              equivalent to selecting Reference Waveforms from the File menu, choosing Label from
              the drop-down list, selecting the tab associated with the reference for which you want
              to create a label, and entering a label in the Label box.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:LABel:NAMe?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:LABel:NAMe?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``REF<x>:LABel:NAMe value`` command.

        **SCPI Syntax:**

        ::

            - REF<x>:LABel:NAMe <QString>
            - REF<x>:LABel:NAMe?

        **Info:**
            - ``<QString>`` is the character string that will be used for the reference waveform
              label name.
        """
        return self._name

    @property
    def xpos(self) -> RefItemLabelXpos:
        """Return the ``REF<x>:LABel:XPOS`` command.

        **Description:**
            - This command sets or queries the X screen offset at which the label (attached to the
              displayed waveform of the specified reference) is displayed, relative to the left edge
              of the screen The reference waveform is specified by x. The value of x can range from
              1 through 4. This command is equivalent to selecting Reference Waveforms from the File
              menu, choosing Label from the drop-down list, selecting the tab associated with the
              reference for which you want to position a label, and entering a value in the X
              Position box.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:LABel:XPOS?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:LABel:XPOS?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``REF<x>:LABel:XPOS value`` command.

        **SCPI Syntax:**

        ::

            - REF<x>:LABel:XPOS <NR1>
            - REF<x>:LABel:XPOS?

        **Info:**
            - ``<NR1>`` is the location (control in divisions) where the waveform label for the
              selected reference is displayed, relative to the left edge of the screen. Arguments
              should be integers ranging from 0 through 10.
        """
        return self._xpos

    @property
    def ypos(self) -> RefItemLabelYpos:
        """Return the ``REF<x>:LABel:YPOS`` command.

        **Description:**
            - This command sets or queries the Y screen offset at which the label (attached to the
              displayed waveform of the specified reference) is displayed, relative to the center of
              the screen. The reference waveform is specified by x. The value of x can range from 1
              through 4. This command is equivalent to selecting Reference Waveforms from the File
              menu, choosing Label from the drop-down list, selecting the tab associated with the
              reference for which you want to position a label, and entering a value in the Y
              Position box.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:LABel:YPOS?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:LABel:YPOS?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``REF<x>:LABel:YPOS value`` command.

        **SCPI Syntax:**

        ::

            - REF<x>:LABel:YPOS <NR1>
            - REF<x>:LABel:YPOS?

        **Info:**
            - ``<NR1>`` is the location (in divisions) where the waveform label for the selected
              reference is displayed, relative to the waveform handle. Arguments should be integers
              ranging from 10 to -10.
        """
        return self._ypos


class RefItemHorizontalPosition(SCPICmdWrite, SCPICmdRead):
    """The ``REF<x>:HORizontal:POSition`` command.

    **Description:**
        - This command sets or queries the horizontal display position of the reference waveform.
          The reference waveform is specified by x. The value of x can range from 1 through 4. This
          command is equivalent to selecting Reference Waveforms from the File menu, choosing
          Reference Setup¼ from the drop-down list, selecting a reference waveform, and then
          entering the horizontal position value using the multipurpose knob.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:HORizontal:POSition?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:HORizontal:POSition?`` query
          and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``REF<x>:HORizontal:POSition value``
          command.

    **SCPI Syntax:**

    ::

        - REF<x>:HORizontal:POSition <NR3>
        - REF<x>:HORizontal:POSition?

    **Info:**
        - ``<NR3>`` argument specifies the horizontal position of the specified reference waveform
          in percentage of the waveform that is displayed to the right of the center vertical
          graticule. The range of this argument is from 0 through 100.
    """


class RefItemHorizontal(SCPICmdRead):
    """The ``REF<x>:HORizontal`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>:HORizontal?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>:HORizontal?`` query and raise
          an AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.position``: The ``REF<x>:HORizontal:POSition`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._position = RefItemHorizontalPosition(device, f"{self._cmd_syntax}:POSition")

    @property
    def position(self) -> RefItemHorizontalPosition:
        """Return the ``REF<x>:HORizontal:POSition`` command.

        **Description:**
            - This command sets or queries the horizontal display position of the reference
              waveform. The reference waveform is specified by x. The value of x can range from 1
              through 4. This command is equivalent to selecting Reference Waveforms from the File
              menu, choosing Reference Setup¼ from the drop-down list, selecting a reference
              waveform, and then entering the horizontal position value using the multipurpose knob.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:HORizontal:POSition?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:HORizontal:POSition?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``REF<x>:HORizontal:POSition value``
              command.

        **SCPI Syntax:**

        ::

            - REF<x>:HORizontal:POSition <NR3>
            - REF<x>:HORizontal:POSition?

        **Info:**
            - ``<NR3>`` argument specifies the horizontal position of the specified reference
              waveform in percentage of the waveform that is displayed to the right of the center
              vertical graticule. The range of this argument is from 0 through 100.
        """
        return self._position


class RefItem(ValidatedDynamicNumberCmd, SCPICmdRead):
    """The ``REF<x>`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``REF<x>?`` query.
        - Using the ``.verify(value)`` method will send the ``REF<x>?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.horizontal``: The ``REF<x>:HORizontal`` command tree.
        - ``.label``: The ``REF<x>:LABel`` command.
        - ``.threshold``: The ``REF<x>:THRESHold`` command.
        - ``.vertical``: The ``REF<x>:VERTical`` command tree.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "REF<x>") -> None:
        super().__init__(device, cmd_syntax)
        self._horizontal = RefItemHorizontal(device, f"{self._cmd_syntax}:HORizontal")
        self._label = RefItemLabel(device, f"{self._cmd_syntax}:LABel")
        self._threshold = RefItemThreshold(device, f"{self._cmd_syntax}:THRESHold")
        self._vertical = RefItemVertical(device, f"{self._cmd_syntax}:VERTical")

    @property
    def horizontal(self) -> RefItemHorizontal:
        """Return the ``REF<x>:HORizontal`` command tree.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:HORizontal?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:HORizontal?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.position``: The ``REF<x>:HORizontal:POSition`` command.
        """
        return self._horizontal

    @property
    def label(self) -> RefItemLabel:
        """Return the ``REF<x>:LABel`` command.

        **Description:**
            - This query-only command returns a branch query containing the waveform label name and
              the coordinates at which the label (attached to the displayed waveform of the
              specified reference) is displayed. The reference waveform is specified by x. The value
              of x can range from 1 through 4. This command is equivalent to selecting Reference
              Waveforms from the File menu and then choosing Label from the drop-down list.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:LABel?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:LABel?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        **SCPI Syntax:**

        ::

            - REF<x>:LABel?

        Sub-properties:
            - ``.name``: The ``REF<x>:LABel:NAMe`` command.
            - ``.xpos``: The ``REF<x>:LABel:XPOS`` command.
            - ``.ypos``: The ``REF<x>:LABel:YPOS`` command.
        """
        return self._label

    @property
    def threshold(self) -> RefItemThreshold:
        """Return the ``REF<x>:THRESHold`` command.

        **Description:**
            - This command sets or queries the comparable threshold for converting the reference
              signal to digital form for the channel specified by x. The value of x can range from 1
              through 4.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:THRESHold?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:THRESHold?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``REF<x>:THRESHold value`` command.

        **SCPI Syntax:**

        ::

            - REF<x>:THRESHold <NR3>
            - REF<x>:THRESHold?

        **Info:**
            - ``<NR3>`` specifies the reference threshold in volts.
        """
        return self._threshold

    @property
    def vertical(self) -> RefItemVertical:
        """Return the ``REF<x>:VERTical`` command tree.

        **Usage:**
            - Using the ``.query()`` method will send the ``REF<x>:VERTical?`` query.
            - Using the ``.verify(value)`` method will send the ``REF<x>:VERTical?`` query and raise
              an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.position``: The ``REF<x>:VERTical:POSition`` command.
            - ``.scale``: The ``REF<x>:VERTical:SCAle`` command.
        """
        return self._vertical



"""The filtervu commands module.

These commands are used in the following models:
DPO2K, DPO2KB, MSO2K, MSO2KB

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - FILTERVu:FREQuency <NR3>
    - FILTERVu:FREQuency:AVAILable?
    - FILTERVu:FREQuency?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class FiltervuFrequencyAvailable(SCPICmdRead):
    """The ``FILTERVu:FREQuency:AVAILable`` command.

    **Description:**
        - Returns a comma separated list of filter frequencies available based on the current
          instrument settings. The source waveform (as specified by the ``DATA:SOURCE``) must be
          turned on for this query to generate a response.

    **Usage:**
        - Using the ``.query()`` method will send the ``FILTERVu:FREQuency:AVAILable?`` query.
        - Using the ``.verify(value)`` method will send the ``FILTERVu:FREQuency:AVAILable?`` query
          and raise an AssertionError if the returned value does not match ``value``.

    **SCPI Syntax:**

    ::

        - FILTERVu:FREQuency:AVAILable?
    """


class FiltervuFrequency(SCPICmdWrite, SCPICmdRead):
    """The ``FILTERVu:FREQuency`` command.

    **Description:**
        - Sets or queries the FilterVu frequency to the closest value supported with the current
          acquisition settings.

    **Usage:**
        - Using the ``.query()`` method will send the ``FILTERVu:FREQuency?`` query.
        - Using the ``.verify(value)`` method will send the ``FILTERVu:FREQuency?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``FILTERVu:FREQuency value`` command.

    **SCPI Syntax:**

    ::

        - FILTERVu:FREQuency <NR3>
        - FILTERVu:FREQuency?

    **Info:**
        - ``<NR3>`` is the integer format for the current FilterVu frequency setting.

    Properties:
        - ``.available``: The ``FILTERVu:FREQuency:AVAILable`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._available = FiltervuFrequencyAvailable(device, f"{self._cmd_syntax}:AVAILable")

    @property
    def available(self) -> FiltervuFrequencyAvailable:
        """Return the ``FILTERVu:FREQuency:AVAILable`` command.

        **Description:**
            - Returns a comma separated list of filter frequencies available based on the current
              instrument settings. The source waveform (as specified by the ``DATA:SOURCE``) must be
              turned on for this query to generate a response.

        **Usage:**
            - Using the ``.query()`` method will send the ``FILTERVu:FREQuency:AVAILable?`` query.
            - Using the ``.verify(value)`` method will send the ``FILTERVu:FREQuency:AVAILable?``
              query and raise an AssertionError if the returned value does not match ``value``.

        **SCPI Syntax:**

        ::

            - FILTERVu:FREQuency:AVAILable?
        """
        return self._available


class Filtervu(SCPICmdRead):
    """The ``FILTERVu`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``FILTERVu?`` query.
        - Using the ``.verify(value)`` method will send the ``FILTERVu?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.frequency``: The ``FILTERVu:FREQuency`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "FILTERVu") -> None:
        super().__init__(device, cmd_syntax)
        self._frequency = FiltervuFrequency(device, f"{self._cmd_syntax}:FREQuency")

    @property
    def frequency(self) -> FiltervuFrequency:
        """Return the ``FILTERVu:FREQuency`` command.

        **Description:**
            - Sets or queries the FilterVu frequency to the closest value supported with the current
              acquisition settings.

        **Usage:**
            - Using the ``.query()`` method will send the ``FILTERVu:FREQuency?`` query.
            - Using the ``.verify(value)`` method will send the ``FILTERVu:FREQuency?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``FILTERVu:FREQuency value`` command.

        **SCPI Syntax:**

        ::

            - FILTERVu:FREQuency <NR3>
            - FILTERVu:FREQuency?

        **Info:**
            - ``<NR3>`` is the integer format for the current FilterVu frequency setting.

        Sub-properties:
            - ``.available``: The ``FILTERVu:FREQuency:AVAILable`` command.
        """
        return self._frequency



# ruff: noqa: D402,PLR0913
# pyright: reportConstantRedefinition=none
"""The scriptvar commands module.

These commands are used in the following models:
SS3706A

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Attributes and Functions:

::

    - scriptVar.autorun
    - scriptVar.list()
    - scriptVar.name
    - scriptVar.run()
    - scriptVar.save()
    - scriptVar.source
"""
from typing import Optional, TYPE_CHECKING, Union

from .._helpers import BaseTSPCmd, NoDeviceProvidedError

if TYPE_CHECKING:
    from tm_devices.drivers.pi.tsp_device import TSPDevice


class Scriptvar(BaseTSPCmd):
    """The ``scriptVar`` command tree.

    **Info:**
        - ``scriptVar``, the name of the variable that references the script.

    Properties/methods:
        - ``.autorun``: The ``scriptVar.autorun`` attribute.
        - ``.list()``: The ``scriptVar.list()`` function.
        - ``.name``: The ``scriptVar.name`` attribute.
        - ``.run()``: The ``scriptVar.run()`` function.
        - ``.save()``: The ``scriptVar.save()`` function.
        - ``.source``: The ``scriptVar.source`` attribute.
    """

    def __init__(self, device: Optional["TSPDevice"] = None, cmd_syntax: str = "scriptVar") -> None:
        super().__init__(device, cmd_syntax)

    @property
    def autorun(self) -> str:
        """Access the ``scriptVar.autorun`` attribute.

        **Description:**
            - This attribute controls the autorun state of a script.

        **Usage:**
            - Accessing this property will send the ``print(scriptVar.autorun)`` query.
            - Setting this property to a value will send the ``scriptVar.autorun = value`` command.

        **TSP Syntax:**

        ::

            - scriptVar.autorun = value
            - print(scriptVar.autorun)

        **Info:**
            - ``scriptVar``, the name of the variable that references the script.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".autorun"
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.autorun)"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.autorun`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @autorun.setter
    def autorun(self, value: Union[str, float]) -> None:
        """Access the ``scriptVar.autorun`` attribute.

        **Description:**
            - This attribute controls the autorun state of a script.

        **Usage:**
            - Accessing this property will send the ``print(scriptVar.autorun)`` query.
            - Setting this property to a value will send the ``scriptVar.autorun = value`` command.

        **TSP Syntax:**

        ::

            - scriptVar.autorun = value
            - print(scriptVar.autorun)

        **Info:**
            - ``scriptVar``, the name of the variable that references the script.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_verification_enabled:  # type: ignore[union-attr]
                self._device.set_and_check(  # type: ignore[union-attr]
                    self._cmd_syntax + ".autorun", value
                )
            else:
                self._device.write(  # type: ignore[union-attr]
                    f"{self._cmd_syntax}.autorun = {value}"
                )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.autorun`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @property
    def name(self) -> str:
        """Access the ``scriptVar.name`` attribute.

        **Description:**
            - This attribute contains the name of a script in the runtime environment.

        **Usage:**
            - Accessing this property will send the ``print(scriptVar.name)`` query.
            - Setting this property to a value will send the ``scriptVar.name = value`` command.

        **TSP Syntax:**

        ::

            - scriptVar.name = value
            - print(scriptVar.name)

        **Info:**
            - ``scriptVar``, the name of the variable that references the script.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".name"
            return self._device.query(f"print({self._cmd_syntax}.name)")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.name`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @name.setter
    def name(self, value: Union[str, float]) -> None:
        """Access the ``scriptVar.name`` attribute.

        **Description:**
            - This attribute contains the name of a script in the runtime environment.

        **Usage:**
            - Accessing this property will send the ``print(scriptVar.name)`` query.
            - Setting this property to a value will send the ``scriptVar.name = value`` command.

        **TSP Syntax:**

        ::

            - scriptVar.name = value
            - print(scriptVar.name)

        **Info:**
            - ``scriptVar``, the name of the variable that references the script.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_verification_enabled:  # type: ignore[union-attr]
                self._device.set_and_check(  # type: ignore[union-attr]
                    self._cmd_syntax + ".name", value
                )
            else:
                self._device.write(f"{self._cmd_syntax}.name = {value}")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.name`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    @property
    def source(self) -> str:
        """Access the ``scriptVar.source`` attribute.

        **Description:**
            - This attribute contains the source code of a script.

        **Usage:**
            - Accessing this property will send the ``print(scriptVar.source)`` query.

        **TSP Syntax:**

        ::

            - print(scriptVar.source)

        **Info:**
            - ``scriptVar``, the name of the variable that references the script that contains the
              source code.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            if self._device.command_syntax_enabled:  # type: ignore[union-attr]
                return self._cmd_syntax + ".source"
            return self._device.query(  # type: ignore[union-attr]
                f"print({self._cmd_syntax}.source)"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to access the ``{self._cmd_syntax}.source`` attribute."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def list(self) -> None:
        """Run the ``scriptVar.list()`` function.

        **Description:**
            - This function generates a script listing.

        **TSP Syntax:**

        ::

            - scriptVar.list()

        **Info:**
            - ``scriptVar``, the name of the variable that references the script.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(f"{self._cmd_syntax}.list()")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.list()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def run(self) -> None:
        """Run the ``scriptVar.run()`` function.

        **Description:**
            - This function runs a script.

        **TSP Syntax:**

        ::

            - scriptVar.run()

        **Info:**
            - ``scriptVar``, the name of the variable that references the script.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            self._device.write(f"{self._cmd_syntax}.run()")  # type: ignore[union-attr]
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.run()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error

    def save(self, filename: Optional[str] = None) -> None:
        """Run the ``scriptVar.save()`` function.

        **Description:**
            - This function saves the script to nonvolatile memory or to a USB flash drive.

        **TSP Syntax:**

        ::

            - scriptVar.save()

        Args:
            filename (optional): A string that contains the file name to use when saving the script
                to a USB flash drive.

        Raises:
            tm_devices.commands.NoDeviceProvidedError: Indicates that no device connection exists.
        """
        try:
            function_args = ", ".join(
                str(x)
                for x in (f'"{filename}"' if filename is not None else None,)
                if x is not None
            )
            self._device.write(  # type: ignore[union-attr]
                f"{self._cmd_syntax}.save({function_args})"
            )
        except AttributeError as error:
            msg = f"No TSPDevice object was provided, unable to run the ``{self._cmd_syntax}.save()`` function."  # noqa: E501
            raise NoDeviceProvidedError(msg) from error



"""The recall commands module.

These commands are used in the following models:
LPD6, MSO2, MSO4, MSO5, MSO5B, MSO5LP, MSO6, MSO6B

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - RECAll:MASK <source file>,MASK<x>
    - RECAll:SESsion <file_path>
    - RECAll:SETUp {FACtory|<file_path>}
    - RECAll:WAVEform <source file>,<destination>
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class RecallWaveform(SCPICmdWrite):
    """The ``RECAll:WAVEform`` command.

    **Description:**
        - This command recalls a stored waveform to a reference memory location.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``RECAll:WAVEform value`` command.

    **SCPI Syntax:**

    ::

        - RECAll:WAVEform <source file>,<destination>

    **Info:**
        - ``<source file>`` is the source file. The file is expected to be located in a directory
          relative to the current working directory (specified by ``FILESYSTEM:CWD``) unless a
          complete path is specified.
        - ``<destination>`` is REF<x> which specifies a reference to create from the recalled
          waveform data file.
    """


class RecallSetup(SCPICmdWrite):
    """The ``RECAll:SETUp`` command.

    **Description:**
        - This command (no query form) returns stored or factory settings to the instrument from a
          copy of the settings stored in memory. This command performs the same function as
          selecting Recall from the File menu, and then choosing the Setup button.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``RECAll:SETUp value`` command.

    **SCPI Syntax:**

    ::

        - RECAll:SETUp {FACtory|<file_path>}

    **Info:**
        - ``FACtory`` restores the factory setup. Performs the same operation as the ``:FACtory``
          command.
        - ``<file_path>`` specifies a location for an instrument setup file. <file path> is a quoted
          string that defines the file name and path. If a file name or path is specified, the file
          is expected to be located in a directory relative to the current working directory
          (specified by ``FILESYSTEM:CWD``) unless a complete path is specified.
    """


class RecallSession(SCPICmdWrite):
    """The ``RECAll:SESsion`` command.

    **Description:**
        - Restores the state of the instrument, including reference waveforms, from a saved session
          file.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``RECAll:SESsion value`` command.

    **SCPI Syntax:**

    ::

        - RECAll:SESsion <file_path>

    **Info:**
        - ``<file_path>`` is the file path that specifies the location of the specified instrument
          session file.
    """


class RecallMask(SCPICmdWrite):
    """The ``RECAll:MASK`` command.

    **Description:**
        - This command recalls a saved mask definition from a Mask File. File suffixes can be xml or
          msk. If the specified mask test already exists the mask associated with that mask test
          will be replaced, otherwise a new mask test is created.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``RECAll:MASK value`` command.

    **SCPI Syntax:**

    ::

        - RECAll:MASK <source file>,MASK<x>

    **Info:**
        - ``MASK<x>`` is the destination mask.
        - ``<source file>`` is the source file. The file is expected to be located in a directory
          relative to the current working directory (specified by ``FILESYSTEM:CWD``) unless a
          complete path is specified.
    """


class Recall(SCPICmdRead):
    """The ``RECAll`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``RECAll?`` query.
        - Using the ``.verify(value)`` method will send the ``RECAll?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.mask``: The ``RECAll:MASK`` command.
        - ``.session``: The ``RECAll:SESsion`` command.
        - ``.setup``: The ``RECAll:SETUp`` command.
        - ``.waveform``: The ``RECAll:WAVEform`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "RECAll") -> None:
        super().__init__(device, cmd_syntax)
        self._mask = RecallMask(device, f"{self._cmd_syntax}:MASK")
        self._session = RecallSession(device, f"{self._cmd_syntax}:SESsion")
        self._setup = RecallSetup(device, f"{self._cmd_syntax}:SETUp")
        self._waveform = RecallWaveform(device, f"{self._cmd_syntax}:WAVEform")

    @property
    def mask(self) -> RecallMask:
        """Return the ``RECAll:MASK`` command.

        **Description:**
            - This command recalls a saved mask definition from a Mask File. File suffixes can be
              xml or msk. If the specified mask test already exists the mask associated with that
              mask test will be replaced, otherwise a new mask test is created.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``RECAll:MASK value`` command.

        **SCPI Syntax:**

        ::

            - RECAll:MASK <source file>,MASK<x>

        **Info:**
            - ``MASK<x>`` is the destination mask.
            - ``<source file>`` is the source file. The file is expected to be located in a
              directory relative to the current working directory (specified by ``FILESYSTEM:CWD``)
              unless a complete path is specified.
        """
        return self._mask

    @property
    def session(self) -> RecallSession:
        """Return the ``RECAll:SESsion`` command.

        **Description:**
            - Restores the state of the instrument, including reference waveforms, from a saved
              session file.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``RECAll:SESsion value`` command.

        **SCPI Syntax:**

        ::

            - RECAll:SESsion <file_path>

        **Info:**
            - ``<file_path>`` is the file path that specifies the location of the specified
              instrument session file.
        """
        return self._session

    @property
    def setup(self) -> RecallSetup:
        """Return the ``RECAll:SETUp`` command.

        **Description:**
            - This command (no query form) returns stored or factory settings to the instrument from
              a copy of the settings stored in memory. This command performs the same function as
              selecting Recall from the File menu, and then choosing the Setup button.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``RECAll:SETUp value`` command.

        **SCPI Syntax:**

        ::

            - RECAll:SETUp {FACtory|<file_path>}

        **Info:**
            - ``FACtory`` restores the factory setup. Performs the same operation as the
              ``:FACtory`` command.
            - ``<file_path>`` specifies a location for an instrument setup file. <file path> is a
              quoted string that defines the file name and path. If a file name or path is
              specified, the file is expected to be located in a directory relative to the current
              working directory (specified by ``FILESYSTEM:CWD``) unless a complete path is
              specified.
        """
        return self._setup

    @property
    def waveform(self) -> RecallWaveform:
        """Return the ``RECAll:WAVEform`` command.

        **Description:**
            - This command recalls a stored waveform to a reference memory location.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``RECAll:WAVEform value`` command.

        **SCPI Syntax:**

        ::

            - RECAll:WAVEform <source file>,<destination>

        **Info:**
            - ``<source file>`` is the source file. The file is expected to be located in a
              directory relative to the current working directory (specified by ``FILESYSTEM:CWD``)
              unless a complete path is specified.
            - ``<destination>`` is REF<x> which specifies a reference to create from the recalled
              waveform data file.
        """
        return self._waveform



"""SMU Model 2401 device driver module."""
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2400_standard import SMU2400Standard


class SMU2401(SMU2400Standard):
    """SMU2401 device driver."""
